/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
**  KOJAK       http://www.fz-juelich.de/jsc/kojak/                        **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2003-2008                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/**
 * @file  epk_mpicom.c
 * 
 * @brief Communicator, group and window management
 */

#include "epk_mpicom.h"
#include "elg_error.h"
#include "esd_def.h"

#include <stdlib.h>
#include <string.h>

/* 
 *-----------------------------------------------------------------------------
 *
 * Communicator management
 *
 *-----------------------------------------------------------------------------
 */

/** @internal
 * maximum amount of concurrently defined communicators per process */
#define EPK_MAX_COMM    50
/** @internal
 * maximum amount of concurrently defines groups per process */
#define EPK_MAX_GROUP   50
/** @internal
 * maximum amount of concurrently defined windows per process */
#define EPK_MAX_WIN     50
/** @internal 
 * maximum amount of concurrently active access or exposure epochs per
 * process */
#define EPK_MAX_WINACC  50

/** @internal
 * structure to hold the \a MPI_COMM_WORLD definition */
struct EpkWorld
{
  MPI_Group group;
  int size;
  int size_grpv;
  int* ranks;
};

struct EpkWorld world;

/* communicator and rank identifiers are (currently) only needed when tracing */
int epk_comm_determination = 0;

/* -- rank translation -- */

/** @internal
 * Translate a rank within a given communicator to its global rank in
 * \a MPI_COMM_WORLD.
 * @param  rank Local rank to be translated
 * @param  comm Local communicator the rank refers to
 * @return Global rank
 */
int epk_rank_to_pe(int rank, MPI_Comm comm)
{
  MPI_Group group;
  int global_rank;
  int inter;

  /* inter-communicators need different call than intra-communicators */
  PMPI_Comm_test_inter(comm, &inter);
  if ( inter )
    PMPI_Comm_remote_group(comm, &group);
  else
    PMPI_Comm_group(comm, &group);

  /* translate rank with respect to \a MPI_COMM_WORLD */
  PMPI_Group_translate_ranks(group, 1, &rank, world.group, &global_rank);
  /* free internal group of input communicator */
  PMPI_Group_free(&group);

  return global_rank;
}

#ifdef HAS_MPI2_1SIDED
/** @internal
 * Translate a rank to its global rank in reference to a window (and its
 * communicator.
 * @param  rank Local rank to be translated
 * @param  win  Window that implicitely defines the communicator rank
 *              refers to.
 * @return Global rank
 */
int epk_win_rank_to_pe(int rank, MPI_Win win)
{
  MPI_Group group;
  int global_rank;
  
  /* get group of communicator associated with input window */
  PMPI_Win_get_group( win, &group );
  /* translate rank with respect to \a MPI_COMM_WORLD */
  PMPI_Group_translate_ranks( group, 1, &rank, world.group, &global_rank );
  /* free internal group of input window */
  PMPI_Group_free( &group );

  return global_rank;
}

/* -- window handling -- */

/** @internal
 * Structure to translate MPI window handles to internal IDs */
struct EpkWin
{
  MPI_Win          win;
  elg_ui4          wid;
};

/** @internal
 * Value of next available window ID */
static int currwid = 0;
/** @internal
 * index of last valid entry in window tracking data structure */
static int last_win = 0;
/** @internal
 * window tracking array */
static struct EpkWin wins[EPK_MAX_WIN];

/** @internal
 * Return the internal id for a given MPI window handle
 * @param  win MPI window handle
 * @return Internal ID for the given window
 */
elg_ui4 epk_win_id(MPI_Win win)
{
  int i = 0;

  while(i < last_win && wins[i].win != win)
    i++;
  
  if (i <= last_win)
    return wins[i].wid;
  else 
    {
      elg_error_msg("epk_win_id: Cannot find window");  
      return -1;
    }
}

/**
 * Create definition record and internal tracking ID.
 * @param win MPI window handle of the window to be tracked
 * @param comm MPI communication handle the window is defined on
 */
void epk_win_create(MPI_Win win, MPI_Comm comm)
{
  if (last_win >= EPK_MAX_WIN) 
    elg_error_msg("Too many windows");

  /* register mpi window definition */
  /* NOTE: MPI_COMM_WORLD is _not_ present in the internal structures,
   * and _must not_ be queried by epk_comm_id */
  esd_def_mpi_win( currwid, comm == MPI_COMM_WORLD ? 0 : epk_comm_id( comm ) ); 

  /* enter win in wins[] arrray */
  wins[last_win].win = win;
  wins[last_win].wid = currwid++;
  
  last_win++;
}

/**
 * Free definition record and internal tracking ID.
 * @param win MPI window handlte of the window to be freed
 */ 
void epk_win_free(MPI_Win win)
{
  if (last_win == 1 && wins[0].win == win)
      last_win = 0;
  else if (last_win > 1)
    {
      int i = 0;
      
      while(i < last_win && wins[i].win != win)
        i++;
      
      if (i < last_win--)
          wins[i] = wins[last_win];
      else
        elg_error_msg("epk_win_free1: Cannot find window");
    }
  else
      elg_error_msg("epk_win_free2: Cannot find window");
}
#endif

/* -- communicator handling -- */

/** @internal
 * structure for communicator tracking
 */
struct EpkComm
{
  MPI_Comm  comm;
  elg_ui4   cid;
};

/** @internal
 * structure for group tracking
 */
struct EpkGroup
{
  MPI_Group group;
  elg_ui4   gid;
  elg_ui4   refcnt;
};

/** @internal
 * Internal ID to use for the next group or communicator
 * @note As groups and communicators are both represented as MPI_COMM
 *       definition records in the trace, we need to use a single counter 
 *       for the id tracking
 * @note \a MPI_COMM_WORLD is hard-coded to 0, thus user-defined
 *       communicator IDs will start at 1*/
static int             currcid = 1;
/** @internal
 * Index to comm array */
static int             last_comm = 0;
/** @internal
 * Index to group array */
static int             last_group = 0;
/** @internal
 * communicator tracking data structure */
static struct EpkComm  comms[EPK_MAX_COMM];
/** @internal
 * group tracking data structure */
static struct EpkGroup groups[EPK_MAX_GROUP];
/** @internal
 * Internal array used for rank translation */
static int* ranks;
/** @internal
 * Internal bitvector used for bitvector creation of new communicators */
static unsigned char* grpv;
/** @internal
 * Internal flag to indicate communicator initialization */
static int comm_initialized = 0;

/** @internal
 * @brief Allocation and initialization of internal data structures 
 * */
void epk_comm_init()
{
  int i;

  /* check, if we already initialized the data structures */
  if ( !comm_initialized )
  {
    comm_initialized = 1;
    /* get group of \a MPI_COMM_WORLD */
    PMPI_Comm_group(MPI_COMM_WORLD, &world.group);
    /* determine the number of MPI processes */
    PMPI_Group_size(world.group, &world.size);
    /* set byte size of bitvector to store \a MPI_COMM_WORLD */
    world.size_grpv = world.size / 8 + (world.size % 8 ? 1 : 0);

    /* initialize translation data structure for \a MPI_COMM_WORLD */
    world.ranks  = calloc(world.size, sizeof(int));
    for (i = 0; i < world.size; i++)
      world.ranks[i] = i;  

    /* allocate translation buffers */
    ranks  = calloc(world.size, sizeof(int));
    grpv = calloc(world.size_grpv, sizeof(unsigned char));
  }
  else
  {
    elg_warning("epk_comm_init: Duplicate call to communicator initialization ignored!\n");
  }
}

/** @internal
 * @brief Finalize communicator tracking
 */
void epk_comm_finalize()
{
  /* free MPI group held internally */
  PMPI_Group_free(&world.group);

  /* free local translation buffers */
  free(world.ranks);
  free(ranks);
  free(grpv);
}

/** @internal
 * @brief Store bitvector mask of an MPI group in internal translation buffer. 
 *        The nth bit is set, when the rank n is part of the group.
 * @param  group MPI group handle */
void epk_group_to_bitvector(MPI_Group group)
{
  int i, size;

  /* 
   * Determine the world rank of each process in group.
   *
   * Parameter #3 is world.ranks here, as we need an array of integers
   * initialized with 0 to n-1, which world.ranks happens to be. 
   */
  PMPI_Group_size(group, &size);
  PMPI_Group_translate_ranks(group, size, world.ranks, world.group, ranks);

  /* initialize grpv */
  memset(grpv, 0, world.size_grpv);

  /* set corresponding bit for each process in group */
  for (i = 0; i < size; i++)
  {
    grpv[ranks[i] / 8] |= (1 << (ranks[i] % 8));
  }
}

/** @internal
 * @brief Start tracking of a given MPI communcator
 * @param comm MPI communicator handle
 */
void epk_comm_create(MPI_Comm comm)
{
  MPI_Group group;

  /* is storage available */
  if (last_comm >= EPK_MAX_COMM) 
    elg_error_msg("Too many communicators");

  /* get group of this communicator */
  PMPI_Comm_group(comm, &group);
  /* create group entry in grpv */
  epk_group_to_bitvector( group );

  /* register mpi communicator definition */
  esd_def_mpi_comm(currcid, world.size_grpv, grpv);

  /* enter comm in comms[] arrray */
  comms[last_comm].comm  = comm;
  comms[last_comm].cid   = currcid++;
  last_comm++;
  
  /* clean up */
  PMPI_Group_free(&group);
}

/** @internal
 * @brief  Stop tracking of an MPI communicator handle
 * @param  comm still valid MPI communicator that is to be freed by
 *         \a MPI_Comm_free
 */
void epk_comm_free(MPI_Comm comm)
{
  const char* message = "You are trying to free a communicator that was " 
                        "not tracked by Scalasca. Please contact the "
                        "Scalasca support team to get help in resolving "
                        "this issue.";

  /* if only one communicator exists, we just need to decrease \a
   * last_comm */
  if (last_comm == 1 && comms[0].comm == comm)
    {
      last_comm = 0;
    }
  /* if more than one communicator exists, we need to search for the
   * entry */
  else if (last_comm > 1)
    {
      int i = 0;
      
      while(i < last_comm && comms[i].comm != comm)
        i++;
      
      if (i < last_comm--)
      {
        /* swap deletion candidate with last entry in the list */
        comms[i] = comms[last_comm];
      }
      else
        elg_error_msg("elg_comm_free1: %s", message);
    }
  else if (epk_comm_determination)
      elg_error_msg("elg_comm_free2: %s", message);
  else
      elg_warning(message);
}

/** @internal
 * @brief  Retrieves the internal id of a given MPI communicator
 * @param  comm MPI communicator
 * @return internal ID of MPI communicator %comm
 */
elg_ui4 epk_comm_id(MPI_Comm comm)
{
  int i = 0;

  while(i < last_comm && comms[i].comm != comm)
    i++;
  
  if (i != last_comm)
    return comms[i].cid;
  else 
    {
      if (comm == MPI_COMM_WORLD)
        {
          elg_warning("epk_comm_id: This function SHOULD NOT be called with MPI_COMM_WORLD");
          return 0;
        }
      else
        {
          elg_error_msg("epk_comm_id: You are using a communicator that was "
                        "not tracked by Scalasca. Please contact the "
                        "Scalasca support team to get help in resolving "
                        "this issue.");
          return -1;
        }
    }
}

/* 
 *-----------------------------------------------------------------------------
 *
 * Group management
 *
 *-----------------------------------------------------------------------------
 */

/** @internal
 * @brief Start tracking of a given MPI group handle
 * @param group MPI group handle to be tracked
 */
void epk_group_create(MPI_Group group)
{
  int i;

  if (last_group >= EPK_MAX_GROUP) 
    elg_error_msg("Too many groups");

  /* check if group already exists */
  if ((i = epk_group_search(group)) == -1)
    {
      /* create group entry in grpv */
      epk_group_to_bitvector(group);
  
      /* register mpi group definition (as communicator) */
      esd_def_mpi_group(currcid, world.size_grpv, grpv);

      /* enter group in groups[] arrray */
      groups[last_group].group  = group;
      groups[last_group].gid    = currcid++;
      groups[last_group].refcnt = 1;
      last_group++;  
    }
  else
    {
      /* count additional reference on group */
      groups[i].refcnt++;
    }
}

/** @internal
 * @brief Stop tracking of a given MPI group handle
 * @param group MPI group handle
 */
void epk_group_free(MPI_Group group)
{
   if (last_group == 1 && groups[0].group == group)
   {
     groups[0].refcnt--;

     if (groups[0].refcnt == 0)
       last_group--;
   }
   else if (last_group > 1)
   {
     int i;

     if ((i = epk_group_search(group)) != -1 )
       {
         /* decrease reference count on entry */
         groups[i].refcnt--;

         /* check if entry can be deleted */
         if (groups[i].refcnt == 0)
         {
           groups[i] = groups[--last_group];
         }
       }
     else
       elg_error_msg("elg_group_free1: Cannot find group");
   }
   else
     elg_error_msg("elg_group_free2: Cannot find group");
   
}

/** @internal
 * @brief Retrieve the internal ID for a givem MPI group handle
 * @param  group MPI group handle
 * @return Internal ID for the given MPI group handle
 */
elg_ui4 epk_group_id(MPI_Group group)
{
  int i = 0;
  
  while ((i < last_group) && (groups[i].group != group))
    i++;
      
  if (i != last_group)
    return groups[i].gid;  
  
    {
      elg_error_msg("Cannot find group");  
      return -1; 
    }
}

/** @internal
 * @brief  Search internal data structures for the entry of a given MPI
 *         group handle.
 * @param  group  MPI group handle
 * @return Index of entry or -1 if entry could not be found
 */
int epk_group_search(MPI_Group group)
{
  int i = 0;
  
  while ((i < last_group) && (groups[i].group != group))
    i++;
      
  if (i != last_group)
    return i;  
  else
    return -1; 
}

/* 
 *-----------------------------------------------------------------------------
 *
 * Window Access Groups -- which window is accessed by what group
 *
 *-----------------------------------------------------------------------------
 */

#ifdef HAS_MPI2_1SIDED
/** @internal
 * exposure epoch */
const elg_ui1 epk_exp_epoch = 0;
/** @internal
 * access epoch */
const elg_ui1 epk_acc_epoch = 1;

/** @internal
 * Entry data structure to track GATS epochs */
struct EpkWinAcc{
  MPI_Win  win;         /* MPI window identifier */
  elg_ui4  gid;         /* EPIK group id */
  elg_ui1  color;       /* byte to help distiguish accesses on same window */
};

/** @internal
 * data structure to track active GATS epochs */
static struct EpkWinAcc waccs[EPK_MAX_WINACC];
/** @internal
 * index of last valid entry */
static int last_wacc = 0;

/** @internal
 * @brief Start tracking of epoch
 * @param win   MPI window handle of related window
 * @param group MPI group handle of related process group
 * @param color type of epoch (exposure=0, access=1)
 */
void epk_winacc_start(MPI_Win win, MPI_Group group, elg_ui1 color) {
  if (last_wacc >= EPK_MAX_WINACC) 
    elg_error_msg("Too many concurrent window accesses");
    
  waccs[last_wacc].win   = win;
  waccs[last_wacc].gid   = epk_group_id(group);
  waccs[last_wacc].color = color;

  last_wacc++;
}

/** @internal
 * @brief End tracking of epoch
 * @param win   MPI window handle of related window
 * @param color Type of epoch (exposure=0, access=1)
 */
void epk_winacc_end(MPI_Win win, elg_ui1 color) {
  int i = 0;
  /* only one window inside wingrp */
  if (last_wacc == 1 
       && waccs[0].win   == win
       && waccs[0].color == color) 
    {
      last_wacc--;
    } 
  else  
    {
      while ((i <= last_wacc) && 
             ((waccs[i].win != win) || (waccs[i].color != color)))
        i++;
      
      if (i != last_wacc)
        {
          last_wacc--;
          waccs[i].win   = waccs[last_wacc].win;
          waccs[i].gid   = waccs[last_wacc].gid;
          waccs[i].color = waccs[last_wacc].color;
        }
      else
        {
          elg_error_msg("Cannot delete window access entry");
        } 
    }
}

/** @internal
 * @brief  Get internal group ID of process group related to an epoch
 * @param  win   MPI window handle of related window
 * @param  color Type of epoch (exposure=0, access=1)
 * @return internal process group id
 */
elg_ui4 epk_winacc_get_gid(MPI_Win win, elg_ui1 color) {
  int i = 0;

  while ((i <= last_wacc) && 
         ((waccs[i].win != win) || (waccs[i].color != color)))
    i++;
      
  if (i != last_wacc)
    {
      return waccs[i].gid;
    }
  else 
    {
      elg_error_msg("Cannot find window access entry");
      return -1;
    }  
}
#endif
