/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
**  KOJAK       http://www.fz-juelich.de/jsc/kojak/                        **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2003-2008                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <libgen.h>

#include "pomp_lib.h"

#include "elg_error.h"

#include "epk_pomp.h"
#include "epk_omplock.h"
#include "epk_ompreg.h"

#include "esd_def.h"
#include "esd_event.h"
#include "esd_run.h"

/*
 * Global variables
 */

int pomp_tracing = 0;
extern elg_ui4 esd_forking_rid;

/*
 *----------------------------------------------------------------
 * EPIK implementation of the C POMP function library
 *----------------------------------------------------------------
 */
   
void POMP_Finalize() {
  static int pomp_finalize_called = 0;

  if ( ! pomp_finalize_called ) {
    pomp_finalize_called = 1;
    esd_close();
    POMP_Off();
  }
}

void POMP_Init() {
  static int pomp_init_called = 0;
  int i;
  elg_ui1 rtype;
  char* rname;
  const char* rdesc;

  if ( ! pomp_init_called ) {
    pomp_init_called = 1;

    esd_open();
    atexit(POMP_Finalize);

    struct EpkRegDescr *rd_data_table = calloc(POMP_MAX_ID, sizeof(struct EpkRegDescr));
    if (!rd_data_table)
        elg_error_msg("POMP_Init: Failed to allocate table for %d regions", POMP_MAX_ID);
    else
        elg_cntl_msg("Got table from OPARI with %d regions", POMP_MAX_ID);

#ifdef ELG_OMP
    /* register wrapper functions for OpenMP API */
    epk_omp_register();

    /* register OMP regions identified by OPARI */
    for (i = 0; i < POMP_MAX_ID; ++i) {
      if ( pomp_rd_table[i] ) {
        char anno_rname[256]; /* annotated region name */
        struct EpkRegDescr* data = &rd_data_table[i];
        struct ompregdescr* r    = pomp_rd_table[i];      
        r->data = data;
	rdesc   = "OMP";

        if ((strcmp(r->name, "region") == 0) || (strcmp(r->name, "function") == 0)) {
            continue; /* USR regions handled after all OMP regions */
        }

        /* -- register file --*/
        data->fid   = esd_def_file(r->file_name);
        data->begln = r->begin_first_line;
        data->endln = r->end_last_line;
	data->sbrid = ELG_NO_ID;

        if (strcmp(r->name, "atomic") == 0) {
          rtype = ELG_OMP_ATOMIC;
          rname = "!$omp atomic";
        } else if (strcmp(r->name, "barrier") == 0) {
          rtype = ELG_OMP_BARRIER;
          rname = "!$omp barrier";
        } else if (strcmp(r->name, "critical") == 0) {
          rtype = ELG_OMP_CRITICAL;
          rname = "!$omp critical";
          sprintf(anno_rname, "%s @%s:%d", "!$omp critical sblock",
                basename(r->file_name), r->begin_first_line+1);
          data->sbrid = esd_def_region(anno_rname,
			               data->fid, r->begin_last_line+1,
                                       r->end_first_line-1, "OMP",
         			       ELG_OMP_CRITICAL_SBLOCK);
	  if ( r->sub_name[0] ) {
            rname = malloc(24 + strlen(r->sub_name));
	    sprintf(rname, "!$omp critical %s", r->sub_name);
	    data->brid = epk_lock_init(r->sub_name);
	  } else {
	    data->brid = epk_lock_init("default");
          }
        } else if (strcmp(r->name, "do") == 0) {
          rtype = ELG_OMP_LOOP;
          rname = "!$omp do";
        } else if (strcmp(r->name, "flush") == 0) {
          rtype = ELG_OMP_FLUSH;
          rname = "!$omp flush";
        } else if (strcmp(r->name, "for") == 0) {
          rtype = ELG_OMP_LOOP;
          rname = "!$omp for";
        } else if (strcmp(r->name, "master") == 0) {
          rtype = ELG_OMP_MASTER;
          rname = "!$omp master";
        } else if (strcmp(r->name, "parallel") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
        } else if (strcmp(r->name, "paralleldo") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
          sprintf(anno_rname, "%s @%s:%d", "!$omp do",
                basename(r->file_name), r->begin_first_line);
          data->sbrid = esd_def_region(anno_rname,  data->fid, data->begln,
                                       data->endln, "OMP", ELG_OMP_LOOP);
        } else if (strcmp(r->name, "parallelfor") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
          sprintf(anno_rname, "%s @%s:%d", "!$omp for",
                basename(r->file_name), r->begin_first_line);
          data->sbrid = esd_def_region(anno_rname,  data->fid, data->begln,
                                       data->endln, "OMP", ELG_OMP_LOOP);
        } else if (strcmp(r->name, "parallelsections") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
          sprintf(anno_rname, "%s @%s:%d", "!$omp sections",
                basename(r->file_name), r->begin_first_line);
          data->sbrid = esd_def_region(anno_rname, data->fid, data->begln,
                                       data->endln, "OMP", ELG_OMP_SECTIONS);
        } else if (strcmp(r->name, "parallelworkshare") == 0)  {
          rtype = ELG_OMP_PARALLEL;
          rname = "!$omp parallel";
          sprintf(anno_rname, "%s @%s:%d", "!$omp workshare",
                basename(r->file_name), r->begin_first_line);
          data->sbrid = esd_def_region(anno_rname, data->fid, data->begln,
                                       data->endln, "OMP", ELG_OMP_WORKSHARE);
        } else if (strcmp(r->name, "sections") == 0) {
          rtype = ELG_OMP_SECTIONS;
          rname = "!$omp sections";
          sprintf(anno_rname, "%s @%s:%d", "!$omp section",
                basename(r->file_name), r->begin_last_line);
          data->sbrid = esd_def_region(anno_rname,
			               data->fid, r->begin_last_line,
                                       r->end_first_line, "OMP",
         			       ELG_OMP_SECTION);
        } else if (strcmp(r->name, "section") == 0) {
          /* NOT DEFINED BY POMP YET */
          /* rtype = ELG_OMP_SECTION; */
          /* rname = "!$omp section"; */
        } else if (strcmp(r->name, "single") == 0) {
          rtype = ELG_OMP_SINGLE;
          rname = "!$omp single";
          sprintf(anno_rname, "%s @%s:%d", "!$omp single sblock",
                basename(r->file_name), r->begin_last_line+1);
          data->sbrid = esd_def_region(anno_rname,
			               data->fid, r->begin_last_line+1,
                                       r->end_first_line-1, "OMP",
         			       ELG_OMP_SINGLE_SBLOCK);
        } else if (strcmp(r->name, "workshare") == 0) {
          rtype = ELG_OMP_WORKSHARE;
          rname = "!$omp workshare";
        } else {
          elg_warning("Unexpected OPARI record: %s/%s", r->name, r->sub_name);
          rtype = ELG_UNKNOWN;
          rname = (char*) r->name;
          rdesc = "UNKNOWN";
        }

        sprintf(anno_rname, "%s @%s:%d", rname,
                basename(r->file_name), r->begin_first_line);
        rname = anno_rname;

        /* -- register region -- */
        data->rid = esd_def_region(rname, data->fid, data->begln,
                                   data->endln, rdesc, rtype);

        if (rtype == ELG_OMP_PARALLEL ||
            rtype == ELG_OMP_LOOP     ||
            rtype == ELG_OMP_SECTIONS ||
            rtype == ELG_OMP_SINGLE   ||
            rtype == ELG_OMP_WORKSHARE) {
          /* -- register implicit barrier -- */
          rname = "!$omp ibarrier";
          sprintf(anno_rname, "%s @%s:%d", rname,
                basename(r->file_name), r->end_last_line);
          data->brid = esd_def_region(anno_rname,
                                      data->fid, data->endln, data->endln,
                                      "OMP", ELG_OMP_IBARRIER);
        } else
          data->brid = ELG_NO_ID;
      }
    }

    /* synthetic region marking end of set of OpenMP regions
       (and thereby identifying the last region expected to fork) */
    esd_forking_rid = esd_def_region("FORKING", esd_def_file("EPIK"),
                                ELG_NO_NUM, ELG_NO_NUM, "EPIK", ELG_UNKNOWN);
    elg_cntl_msg("esd_forking_rid = %u", esd_forking_rid);
#endif /* ELG_OMP */

    /* register (remaining) POMP regions identified by OPARI */
    for (i = 0; i < POMP_MAX_ID; ++i) {
      if ( pomp_rd_table[i] ) {
        struct EpkRegDescr* data = &rd_data_table[i];
        struct ompregdescr* r    = pomp_rd_table[i];      
        r->data = data;

        if (data->rid != 0) continue; /* already defined */

        if (strcmp(r->name, "region") == 0)  {
          rtype = ELG_USER_REGION;
          rname = (char*) r->sub_name;
	  rdesc   = "USR";
        } else if (strcmp(r->name, "function") == 0)  {
          rtype = ELG_FUNCTION;
          rname = (char*) r->sub_name;
	  rdesc   = "USR";
        } else { /* not a USR region */
          elg_warning("Unexpected OPARI record: %s/%s", r->name, r->sub_name);
          rtype = ELG_UNKNOWN;
          rname = (char*) r->name;
          rdesc = "UNKNOWN";
        }

        /* -- register file --*/
        data->fid   = esd_def_file(r->file_name);
        data->begln = r->begin_first_line;
        data->endln = r->end_last_line;

        /* -- register region -- */
        data->rid = esd_def_region(rname, data->fid, data->begln,
                                   data->endln, rdesc, rtype);
        data->brid  = ELG_NO_ID;
	data->sbrid = ELG_NO_ID;
      }
    }

    POMP_On();
  }
}

void POMP_Off() {
  pomp_tracing = 0;
}

void POMP_On() {
  pomp_tracing += 1;
}

void POMP_Begin(struct ompregdescr* r) {
  struct EpkRegDescr* data = (struct EpkRegDescr*)(r->data);

  if ( pomp_tracing ) esd_enter(data->rid);
}

void POMP_End(struct ompregdescr* r) {
  struct EpkRegDescr* data = (struct EpkRegDescr*)(r->data);

  if ( pomp_tracing ) esd_exit(data->rid);
}
