/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
**  KOJAK       http://www.fz-juelich.de/jsc/kojak/                        **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2003-2008                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

/* XXXX EsdProc should manage EsdThrd objects */

#include "esd_proc.h"
#include "esd_gen.h"
#include "epk_archive.h"
#include "epk_idmap.h"
#include "esd_unify.h"
#include "esd_paths.h"
#include "esd_report.h"

#include "elg_pform.h"
#include "elg_error.h"
#include "epk_conf.h"

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#if (defined (ELG_MPI) || defined (ELG_OMPI))
#include <mpi.h>
#include "epk_mpireg.h"

EXTERN MPI_Comm esd_comm_world;
#endif
#define PATH_TYPE__NONE 0
#define PATH_TYPE__OMP  8 /* XXXX OMP_HACK */


/* number of performance metrics (excluding time) */
EXTERN int esd_num_metrics;

typedef struct 
{
    EsdPaths_t* paths;          /* callpaths management object */
} ElgThrd;
/* vector of thread objects */
EXTERN ElgThrd** thrdv;

EsdProc* EsdProc_open( void )
{
  size_t bsize = epk_str2size(epk_get(ESD_BUFFER_SIZE));
  char tmp[PATH_MAX];
  EsdProc *proc;

  proc = (EsdProc*) malloc(sizeof(EsdProc));
  if ( proc == NULL )
    elg_error();

  /* temporary file basename for local process definitions */  
  sprintf(tmp, "%s/%s.%lX.%u", epk_get(EPK_LDIR), epk_get(EPK_TITLE),
                               elg_pform_node_id(), getpid());
    
  proc->gen = EsdGen_open(tmp, ELG_NO_ID, bsize);

  proc->arc_name = NULL; /* not known yet */
  proc->tmp_name = (char*)calloc(strlen(tmp) + 1, sizeof(char*));
  if ( proc->tmp_name == NULL )
    elg_error();

  strcpy(proc->tmp_name, EsdGen_get_name(proc->gen));

  elg_cntl_msg("Process object created");
  
  return proc;
}

static int move_file(char* ifile, char *ofile)
{ /* rename(2) can only be used for files on the same filesystem */
  int ch, ret;
  FILE *fi = fopen(ifile, "r");
  FILE *fo = fopen(ofile, "w");
  if (!fi) elg_error_msg("Failed to open %s for reading", ifile);  
  if (!fo) elg_error_msg("Failed to open %s for writing", ofile);  
  while (((ch = getc(fi))!=EOF) && ((ret=putc(ch,fo))!=EOF)) /*done*/;
  if (ch!=EOF) elg_error_msg("Failed to write to %s", ofile);
  if (ret==EOF) elg_error_msg("Failed to read from %s", ifile);
  if (fclose(fi)) elg_error_msg("Failed to close %s", ifile);
  if (fclose(fo)) elg_error_msg("Failed to close %s", ofile);
  /* could check ELG_CLEAN do decide whether to remove file */
  if (unlink(ifile)) elg_error_msg("Failed to remove %s", ifile);
  return 0;
}


void EsdProc_close(EsdProc* proc, elg_ui4 rank)
{
  EsdGen_close(proc->gen);

  if (!proc->arc_name) {
      /* destination for process definitions */
      proc->arc_name = epk_archive_rankname(EPK_TYPE_ESD, epk_archive_get_name(), rank);
      epk_archive_create(EPK_TYPE_ESD);

      /* move process definitions file to destination */
      if (move_file(proc->tmp_name, proc->arc_name) == 0)
          elg_cntl_msg("Moved definitions file %s to %s",
                        proc->tmp_name, proc->arc_name);
      else
          elg_error_msg("Failed to move definitions file %s to %s",
                        proc->tmp_name, proc->arc_name);
  }

  elg_cntl_msg("Process object deleted");
  free(proc->tmp_name);
  free(proc->arc_name);
  free(proc);
}


void EsdProc_writedefs (EsdUnifyData* udata)
{
  /* Open global definitions file */
  char*   dirname = epk_archive_get_name();
  char*   outname = epk_archive_filename(EPK_TYPE_ESD, dirname);
  ElgOut* outfile = ElgOut_open(outname, ELG_BYTE_ORDER, ELG_UNCOMPRESSED);
  if (!outfile)
    elg_error_msg("Could not create file \"%s\".", outname);

  /* Write definition data & close file */
  esd_unify_write_defs(udata, outfile);
  ElgOut_close(outfile);

  /* Release resources */
  free(outname);
}

ElgOut* EsdProc_openmap ()
{
  static char* mapfile = "epik.map";
  char* dirname;
  char* filename;
  ElgOut* outfile;

  /* Build filename */
  dirname  = epk_archive_get_name();
  filename = malloc(strlen(dirname) + strlen(mapfile) + 2);
  sprintf(filename, "%s/%s", dirname, mapfile);

  /* Open mapping file */
  outfile = ElgOut_open(filename, ELG_BYTE_ORDER, ELG_UNCOMPRESSED);
  if (!outfile)
    elg_error_msg("Could not create file \"%s\".", filename);

  /* Release memory */
  free(filename);

  return outfile;
}

EXTERN int esd_summary;
EXTERN int esd_tracing;

/* local->global callpath identifier map vector */
static size_t   esd_idmapsz = 0;
static elg_ui4* esd_idmapv = NULL;
static elg_ui4  esd_total_paths = 0;
static elg_ui4  esd_unknowns = 0;

/* vectors of callpath types (MPI collective and pt2pt) */
static elg_ui1* esd_pathtypev = NULL;

/* determine which callpaths have MPI metrics of various kinds */
static void determine_pathtypes(EsdUnifyData* udata)
{
    /* determine MPI & OMP region descid */
    elg_ui4 mpi_descid, omp_descid;
    for (mpi_descid=0; mpi_descid<epk_vector_size(udata->strings); mpi_descid++) {
        if (strcmp(epk_vector_at(udata->strings,mpi_descid), "MPI") == 0) break;
    }
    for (omp_descid=0; omp_descid<epk_vector_size(udata->strings); omp_descid++) {
        if (strcmp(epk_vector_at(udata->strings,omp_descid), "OMP") == 0) break;
    }
    elg_cntl_msg("MPI region descid=%u OMP region descid=%u", mpi_descid, omp_descid);

    /* determine MPI collective and pt2pt callpaths */
    esd_pathtypev = calloc(esd_total_paths, sizeof(elg_ui1));
    if (esd_pathtypev) {
        elg_ui4 pathid;
        elg_ui4 colls=0, pt2pt=0, other=0, pregs=0;
        for (pathid=0; pathid<esd_total_paths; pathid++) {
            UnifyCpath*  ucpath = epk_vector_at(udata->cpaths, pathid);
            UnifyRegion* uregion = epk_vector_at(udata->regions, ucpath->rid);
            if (uregion->descid == omp_descid) { /* XXXX OMP_HACK */
                const char* omp_parallel = "!$omp parallel";
                char* regnname = epk_vector_at(udata->strings, uregion->nameid);
                if (strncmp(regnname, omp_parallel, 14) == 0) {
                    esd_pathtypev[pathid] = PATH_TYPE__OMP;
                    elg_cntl_msg("cpath[%d]: %s eventtype=OMP", pathid, regnname);
                    pregs++; /* OMP paths to parallel regions */
                }
#if (defined (ELG_MPI) || defined(ELG_OMPI))
            } else if (uregion->descid == mpi_descid) {
                char* regnname = epk_vector_at(udata->strings, uregion->nameid);
                elg_ui1 eventtype = epk_mpi_eventtype(regnname);
                elg_cntl_msg("cpath[%d]: %s eventtype=%u", pathid, regnname, eventtype);
                esd_pathtypev[pathid] = eventtype;
                if (eventtype == EPK_MPI_TYPE__NONE)
                    other++; /* MPI paths without additional metrics */
                else if (eventtype == EPK_MPI_TYPE__COLLECTIVE)
                    colls++; /* MPI paths with collective metrics */
                else
                    pt2pt++; /* MPI paths with send and/or receive metrics */
#endif
            } else esd_pathtypev[pathid] = PATH_TYPE__NONE;
        }
        elg_cntl_msg("Got %d collective %d pt2pt %d other MPI callpaths",
                    colls, pt2pt, other);
        elg_cntl_msg("Got %d OMP parallel region callpaths", pregs);
    }
}

int EsdProc_unify (EsdProc* proc, elg_ui4 my_rank, elg_ui4 num_ranks, int sync)
{
  static unsigned Unifier = 0; /* default unification on master */
  double start_time;
  size_t bsize = epk_str2size(epk_get(ESD_BUFFER_SIZE));
  size_t reqbytes = EsdGen_get_bytes(ESDPROC_GEN(proc));
  size_t defbytes = EsdGen_get_held(ESDPROC_GEN(proc));
  buffer_t defbuf = EsdGen_get_data(ESDPROC_GEN(proc));
  size_t maxbytes = reqbytes;
  elg_ui4 frames = esd_frame_depth(thrdv[0]->paths);
  elg_ui4 maxframes = frames;
  elg_ui4 unknowns = esd_path_visits(thrdv[0]->paths, 0);
  elg_ui4 lpaths = esd_paths_size(thrdv[0]->paths);
  elg_ui4 maxpaths = lpaths;
  esd_unknowns = unknowns;

  if (my_rank == Unifier) {
      elg_cntl_msg("Unifying...");
      /*if (esd_tracing) epk_archive_create(EPK_TYPE_ELU);*/
      start_time = elg_pform_wtime();
  }

#if (defined (ELG_MPI) || defined(ELG_OMPI))
  PMPI_Allreduce(&reqbytes, &maxbytes, 1, MPI_UNSIGNED, MPI_MAX, esd_comm_world);
  PMPI_Reduce(&lpaths, &maxpaths, 1, MPI_UNSIGNED, MPI_MAX, Unifier, esd_comm_world);
  PMPI_Reduce(&frames, &maxframes, 1, MPI_UNSIGNED, MPI_MAX, Unifier, esd_comm_world);
  PMPI_Allreduce(&unknowns, &esd_unknowns, 1, MPI_UNSIGNED, MPI_SUM, esd_comm_world);
#endif

  if (maxbytes > bsize) {
      esd_summary = 0; /* disable (further) summarisation */
      if (my_rank == Unifier) {
          elg_warning("Definitions lost due to insufficient buffer capacity!");
          elg_warning("ESD_BUFFER_SIZE needs to be at least %d bytes", maxbytes);
	  elg_warning("Unifying... aborted!");
      }
      return 0;
  } else
      if (my_rank == Unifier)
          elg_cntl_msg("Largest definitions buffer %d bytes", maxbytes);

  if (my_rank == Unifier) { /* rank responsible for unification */
    unsigned rank = 0;
    EsdUnifyData* udata = esd_unify_create();
    ElgOut* mapfile = esd_tracing ? EsdProc_openmap() : NULL;

    for (rank = 0; rank < num_ranks; rank++) {
#if (defined (ELG_MPI) || defined(ELG_OMPI))
        if (rank != Unifier) { /* receive transfered definitions */
            int count;
            MPI_Status stat;
            PMPI_Send(NULL, 0, MPI_INT, rank, 0, esd_comm_world);
            elg_cntl_msg("Signalled rank %d ...", rank);
#if defined(ESD_2ND_DEFBUF)
            if (rank == 1) { /* XXXX should reuse original defbuf! */
                defbuf = (buffer_t) malloc(bsize);
                if (!defbuf) elg_error_msg("Failed to malloc recv buffer");
            }
#endif
            PMPI_Recv(defbuf, bsize, MPI_BYTE, rank, MPI_ANY_TAG,
                                        esd_comm_world, &stat);
            PMPI_Get_count(&stat, MPI_BYTE, &count);
            defbytes = count;
        }
#endif
        elg_cntl_msg("Unifying rank %d (%d bytes) ...", rank, defbytes);
        esd_unify_buffer(udata, defbuf, defbytes, mapfile);

        const EpkIdMap* idmap = esd_unify_get_map(udata, ESD_UNIFY_CPATHMAP);
        const elg_ui4* idmapv = epk_idmap_data(idmap, &esd_idmapsz); /* DENSE */
        size_t mapbytes = esd_idmapsz * sizeof(elg_ui4);
        if (rank == Unifier) {
            if (esd_idmapsz < lpaths)
                elg_warning("Only %d/%d unifier callpaths retained", esd_idmapsz, lpaths);
            esd_idmapv = malloc(mapbytes);
            memcpy(esd_idmapv, idmapv, mapbytes); /* save unifier map */
        }
        elg_cntl_msg("idmap for rank %d of size %d (%d bytes) ...",
                rank, esd_idmapsz, mapbytes);
#if (defined (ELG_MPI) || defined(ELG_OMPI))
        if (esd_summary && (rank != Unifier)) { /* return unified callpath idmap */
            PMPI_Send((void*)idmapv, esd_idmapsz, MPI_UNSIGNED, rank, 0, esd_comm_world);
        }
#endif
        esd_unify_clear_maps(udata);
    }

#if defined(ESD_2ND_DEFBUF)
    if (num_ranks > 1) free(defbuf); /* finished with temporary buffer */
#endif

    if (esd_tracing) { /* complete writing map and defs files */
        ElgOut_close(mapfile);
        EsdProc_writedefs(udata);
    }

    esd_total_paths = epk_vector_size(udata->cpaths);
    elg_warning("%d unique paths (%d max paths, %d max frames, %d unknowns)",
                    esd_total_paths, maxpaths, maxframes, esd_unknowns);
    if (esd_unknowns) elg_warning("Hint: Increase ESD_PATHS to avoid/reduce unknown paths");

    esd_report_paths(udata, esd_unknowns); /* generate paths report */

    if (esd_summary) {
        determine_pathtypes(udata);
        /* open report and write unified definitions */
        if (esd_report_open(NULL)) esd_report_defs(udata, esd_unknowns);
    }

    esd_unify_free(udata); /* release unification data object */
  }
#if (defined (ELG_MPI) || defined(ELG_OMPI))
  else { /* send buffer to Unifier */
    MPI_Status stat;
    PMPI_Recv(NULL, 0, MPI_INT, Unifier, 0, esd_comm_world, &stat);
    elg_cntl_msg("Transfering %d bytes of definitions to master", defbytes);
    PMPI_Send(defbuf, defbytes, MPI_BYTE, Unifier, 0, esd_comm_world);

    if (esd_summary) {
        /* receive idmap from Unifier */
        int count;
        elg_cntl_msg("Awaiting idmap for %u callpaths...", lpaths);
        PMPI_Probe(Unifier, 0, esd_comm_world, &stat);
        PMPI_Get_count(&stat, MPI_UNSIGNED, &count);
        esd_idmapsz = count;
        esd_idmapv = calloc(esd_idmapsz, sizeof(elg_ui4));
        PMPI_Recv(esd_idmapv, esd_idmapsz, MPI_UNSIGNED, Unifier, 0, esd_comm_world, &stat);
        if (esd_idmapsz != lpaths)
            elg_warning("Got callpath idmap size %d instead of %d returned", esd_idmapsz, lpaths);
      }
    }
#endif  

#if (defined (ELG_MPI) || defined(ELG_OMPI))
    /* ensure all processes know number of unique paths */
    PMPI_Bcast(&esd_total_paths, 1, MPI_UNSIGNED, 0, esd_comm_world);
    /* NB: pseudo-barrier since Unifier rank 0 is busy until all ranks are unified */
    /* (prevents interference with subsequent trace flushing, if no summary collation) */
#endif

    if (my_rank == Unifier)
        elg_warning("Unifying... done (%0.3fs)", elg_pform_wtime()-start_time);
    return 1;
}

void EsdProc_collate (EsdProc* proc, elg_ui4 my_rank, elg_ui4 num_ranks, elg_ui4 num_thrds)
{
    unsigned t;
    static unsigned Unifier = 0; /* master */
    double start_time = elg_pform_wtime();
    elg_ui4 local_paths = esd_paths_size(thrdv[0]->paths);
    double* sevs = calloc(num_thrds, sizeof(double)); /* sev for proc threads */
    double* sevbuf = calloc(num_ranks*num_thrds, sizeof(double));
    int ready = (sevs && sevbuf);

    for (t=0; t<num_thrds; t++) {
        thrdv[t]->paths->rev_idmapv = calloc(esd_total_paths+1, sizeof(elg_ui4));
        if (!thrdv[t]->paths->rev_idmapv) ready=0;
    }

    elg_cntl_msg("Collating report (%d/%d paths)...", local_paths, esd_total_paths);

#if (defined (ELG_MPI) || defined(ELG_OMPI))
    if (my_rank != Unifier) esd_pathtypev = calloc(esd_total_paths, sizeof(elg_ui1));
    ready = (ready && esd_pathtypev);
#endif

    /* verify readiness of all ranks to collate report */
    if (!ready) elg_warning("Failed to allocate collation storage!");
    int allready = ready;
#if (defined (ELG_MPI) || defined(ELG_OMPI))
    PMPI_Allreduce(&ready, &allready, 1, MPI_INT, MPI_SUM, esd_comm_world);
#endif
    if (allready != num_ranks) return; /* no report possible */

#if (defined (ELG_MPI) || defined(ELG_OMPI))
    /* distribute callpath type vector */
    PMPI_Bcast(esd_pathtypev, esd_total_paths, MPI_BYTE, Unifier, esd_comm_world);
#endif

    int gather = (getenv("ESD_GATHER_ALL") != NULL);
    if (my_rank == Unifier) elg_cntl_msg("ESD_GATHER_ALL=%d", gather);

    /* initialise reverse map global->local (for master thread) */
    unsigned localid;
    for (localid = 0; localid < local_paths; localid++) {
        elg_ui4 unifyid = esd_idmapv[localid];
        thrdv[0]->paths->rev_idmapv[unifyid] = localid;
        if (unifyid != localid)
            elg_cntl_msg("cpath idmapv[%d] = %d", localid, unifyid);
    }
    /* initialise reverse map global->local for worker threads */
    for (t=1; t<num_thrds; t++) {
        elg_ui4 local_paths = esd_paths_size(thrdv[t]->paths);
        for (localid = 0; localid < local_paths; localid++) {
            elg_ui4 masterid = thrdv[t]->paths->masterv[localid];
            elg_ui4 unifyid = esd_idmapv[masterid];
            thrdv[t]->paths->rev_idmapv[unifyid] = localid;
            elg_cntl_msg("t%d[%d]->t0[%d]->U[%d]", t, localid, masterid, unifyid);
        }
    }

    unsigned unknown = (esd_unknowns) ? 0 : 1;
    unsigned metid, pathid;
#if defined(MSG_HACK)
    const unsigned MPI_METRICS=12, OMP_METRICS=2;
    const unsigned HACK_METRICS=MPI_METRICS+OMP_METRICS;
#else
    const unsigned HACK_METRICS=0;
#endif
    for (metid = 0; metid < esd_num_metrics+2+HACK_METRICS; metid++) {
        for (pathid = unknown; pathid < esd_total_paths; pathid++) {
#if defined(MSG_HACK)
#if (defined (ELG_MPI) || defined(ELG_OMPI))
            /* no need to gather MPI statistics from non-MPI callpaths */
            if (!gather) switch (metid) {
                case  1: /* coll_syncs */
                case  4: /* coll_c_xch */
                case  5: /* coll_c_src */
                case  6: /* coll_c_dst */
                case  9: /* coll_bsent */
                case 10: /* coll_brcvd */
                    /* skip to next pathid with collective metric data */
                    while ((pathid < esd_total_paths) &&
                           !(esd_pathtypev[pathid]&EPK_MPI_TYPE__COLLECTIVE))
                        pathid++;
                    if (pathid == esd_total_paths) continue;
                    break;
                case  2: /* send_syncs */
                case  7: /* send_comms */
                case 11: /* send_bytes */
                    /* skip to next pathid with send metric data */
                    while ((pathid < esd_total_paths) &&
                           !(esd_pathtypev[pathid]&EPK_MPI_TYPE__SEND))
                        pathid++;
                    if (pathid == esd_total_paths) continue;
                    break;
                case  3: /* recv_syncs */
                case  8: /* recv_comms */
                case 12: /* recv_bytes */
                    /* skip to next pathid with recv metric data */
                    while ((pathid < esd_total_paths) &&
                           !(esd_pathtypev[pathid]&EPK_MPI_TYPE__RECV))
                        pathid++;
                    if (pathid == esd_total_paths) continue;
                    break;
                default:
                    break;
            }
#else
            /* no need to gather MPI statistics from non-MPI programs */
            if (!gather && ((metid >= 1) && (metid <= MPI_METRICS))) continue;
#endif
            /* XXXX OMP_HACK */
            if ((metid == 13) || (metid == 14)) {
                /* skip to next OMP pathid */
                while ((pathid < esd_total_paths) &&
                       !(esd_pathtypev[pathid]&PATH_TYPE__OMP))
                    pathid++;
                if (pathid == esd_total_paths) continue;
                elg_cntl_msg("Collating time for OMP path %d", pathid);
            }
#endif
            for (t=0; t<num_thrds; t++) { /* for each thread on process */
                /* determine local path identifier */
                unsigned lpathid = thrdv[t]->paths->rev_idmapv[pathid];
                /* extract this path's severity value */
                if (metid == 0)
                    sevs[t] = esd_path_visits(thrdv[t]->paths, lpathid);
                else if (metid == 1)
#if defined(MSG_HACK)
                /* hack for message statistics: XXX check range above XXX */
                    sevs[t] = esd_path_coll_syncs(thrdv[t]->paths, lpathid);
                else if (metid == 2)
                    sevs[t] = esd_path_send_syncs(thrdv[t]->paths, lpathid);
                else if (metid == 3)
                    sevs[t] = esd_path_recv_syncs(thrdv[t]->paths, lpathid);
                else if (metid == 4)
                    sevs[t] = esd_path_coll_c_xch(thrdv[t]->paths, lpathid);
                else if (metid == 5)
                    sevs[t] = esd_path_coll_c_src(thrdv[t]->paths, lpathid);
                else if (metid == 6)
                    sevs[t] = esd_path_coll_c_dst(thrdv[t]->paths, lpathid);
                else if (metid == 7)
                    sevs[t] = esd_path_send_comms(thrdv[t]->paths, lpathid);
                else if (metid == 8)
                    sevs[t] = esd_path_recv_comms(thrdv[t]->paths, lpathid);
                else if (metid == 9)
                    sevs[t] = esd_path_coll_bsent(thrdv[t]->paths, lpathid);
                else if (metid == 10)
                    sevs[t] = esd_path_coll_brcvd(thrdv[t]->paths, lpathid);
                else if (metid == 11)
                    sevs[t] = esd_path_send_bytes(thrdv[t]->paths, lpathid);
                else if (metid == 12)
                    sevs[t] = esd_path_recv_bytes(thrdv[t]->paths, lpathid);
                else if (metid == 13)
/* XXXX OMP_HACK */
                    sevs[t] = esd_path_tfork_time(thrdv[t]->paths, lpathid);
                else if (metid == 14)
                    sevs[t] = esd_path_tmgmt_time(thrdv[t]->paths, lpathid);
                else if (metid == (HACK_METRICS+1))
#endif
                    sevs[t] = esd_path_timing(thrdv[t]->paths, lpathid);
                else
                    sevs[t] = esd_path_metric(thrdv[t]->paths, lpathid, metid-2-HACK_METRICS);
                sevbuf[t] = sevs[t];
            }

#if (defined (ELG_MPI) || defined(ELG_OMPI))
            elg_cntl_msg("Gather met=%d path=%d", metid, pathid);
            PMPI_Gather(sevs, num_thrds, MPI_DOUBLE,
                        sevbuf, num_thrds, MPI_DOUBLE, Unifier, esd_comm_world);
            PMPI_Barrier(esd_comm_world); /* hold back network flood */
#endif

            if (my_rank == Unifier) { /* add severities to report */
#if DEBUG
                elg_ui4 total = 0;
                if (metid == 0) {
                    for (unsigned r=0; r<num_ranks; r++) {
                        elg_ui4 visits = sevbuf[r];
                        total += visits;
                        printf("P%03u:visits[%u]=%u\n", pathid, r, visits);
                    }
/*                  printf("P%03u:visits=%u\n", pathid, total);*/
                    esd_report_visits(pathid-unknown, total);
                }
#endif
                esd_report_sev_threads(metid, pathid-unknown, sevbuf);
            }
        } /* pathid */
    } /* metid */

#if (defined (ELG_MPI) || defined(ELG_OMPI))
    PMPI_Barrier(esd_comm_world); /* XXXX temp sync */
#endif

    if (my_rank == Unifier) {
        esd_report_close();
        elg_warning("Collating... done (%0.3fs)", elg_pform_wtime()-start_time);
    }
}
