/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
**  KOJAK       http://www.fz-juelich.de/jsc/kojak/                        **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2003-2008                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


#include <assert.h>
#include <stdlib.h>
#include <string.h>

#include "elg_error.h"
#include "elg_readcb.h"
#include "epk_hashtab.h"
#include "epk_vector.h"
#include "esd_unify.h"


/*--- Header file documentation -------------------------------------------*/

/**
 * @addtogroup EPISODE_unify
 * @{
 */
/**
 * @file  esd_unify.h
 * @brief Static program entity unification routines
 *
 * This file contains type definitions and functions prototypes needed for the
 * unification of per-process EPILOG definition records for static program
 * entities, such as instrumented source code regions or MPI communicators.
 *
 * The basic usage model of this module is as follows: Before any data can
 * be processed, an unification data structure has to be created.
 * @code
 *   EsdUnifyData* data = esd_unify_create();
 * @endcode
 * Now, the EPILOG definition data can be processed by calling
 * esd_unify_buffer() for the data generated by each process. Note that the
 * current implementation requires these calls to be <b>in ascending rank
 * order</b>. The unification process also collects per-process information,
 * such as mappings from local to global identifiers or the numbers of events
 * of the different types, which is stored by the esd_unify_buffer() call
 * to a global mapping file.
 *
 * To be able to access these mapping tables afterwards, they are not cleared
 * automatically, i.e., this has to be done manually before processing the data
 * of another process.
 * @code
 *   for (i = 0; i < num_procs; i++) {
 *     esd_unify_buffer(data, buffer[i], bufsize[i], mapfile);
 *     esd_unify_clear_maps(data);
 *   }
 * @endcode
 * After all definitions have been processed, the unified definitions can be
 * written to a global definition file and the unification data structure can
 * be released.
 * @code
 *    esd_unify_write_defs(data, defsfile);
 *    esd_unify_free(data);
 * @endcode
 *
 * @note This module uses the @c assert() macro to check various conditions
 *       (especially the values of given parameters) at runtime, which can
 *       cause a performance penalty.
 */
/** @} */


/*
 *---------------------------------------------------------------------------
 * Local stuff
 *---------------------------------------------------------------------------
 */

/* Default sizes of data structures */
#define TMP_COORD_ARRAY         4
#define STRINGS_HASH         5077
#define STRINGS_ARRAY         512
#define MACHS_ARRAY             4
#define NODES_ARRAY            64
#define PROCESS_ARRAY          16
#define THREADS_ARRAY          16
#define FILES_ARRAY            16
#define LOCS_ARRAY           1024
#define REGIONS_ARRAY        1024
#define COMS_ARRAY             16
#define CSITES_ARRAY           16
#define CPATHS_ARRAY          256
#define METRICS_ARRAY          16
#define TOPOLOGIES_ARRAY        4
#define WINDOWS_ARRAY          16
#define OFFSET_ARRAY         1024


/**
 * @brief  Lookup table for bit count. 
 *
 * The lookup table includes the number of bits set in each possible
 * value of a byte. It could be algorithmically initialised with:
 *
 * @code
 * BitsSetTable[0] = 0;
 * for (int i = 0; i < 256; i++)
 * {
 *   BitsSetTable[i] = (i & 1) + BitsSetTable[i / 2];
 * }
 * @endcode
 */
static const unsigned char BitSetTable[] =
{
  0, 1, 1, 2, 1, 2, 2, 3, 1, 2, 2, 3, 2, 3, 3, 4, 
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
  1, 2, 2, 3, 2, 3, 3, 4, 2, 3, 3, 4, 3, 4, 4, 5, 
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
  2, 3, 3, 4, 3, 4, 4, 5, 3, 4, 4, 5, 4, 5, 5, 6, 
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
  3, 4, 4, 5, 4, 5, 5, 6, 4, 5, 5, 6, 5, 6, 6, 7, 
  4, 5, 5, 6, 5, 6, 6, 7, 5, 6, 6, 7, 6, 7, 7, 8
};

/*--- Function prototypes -------------------------------------------------*/

/* Hash functions */
static size_t hash_string(const char* key);

/* Unification functions */
static void unify_string(EsdUnifyData* data, char* key, elg_ui4 local_id);

/* Comparisons  */
static int single_id_cmp(const UnifySingleId* key, const UnifySingleId* entry);
static int double_id_cmp(const UnifyDoubleId* key, const UnifyDoubleId* entry);
static int region_cmp   (const UnifyRegion*   key, const UnifyRegion*   entry);
static int csite_cmp    (const UnifyCsite*    key, const UnifyCsite*    entry);
static int com_cmp      (const UnifyCom*      key, const UnifyCom*      entry);
static int grpcnt_cmp   (const GrpCount*      key, const GrpCount*      entry);

/* Memory management */
static void strhash_free(EpkHashEntry*  entry);
static void cpath_free  (UnifyCpath*    entry);
static void com_free    (UnifyCom*      entry);
static void cart_free   (UnifyCart*     entry);
static void coord_free  (CartCoord*     entry);
static void grpcnt_free (GrpCount*      entry);

static void grpcnt_reset(GrpCount*      entry);

/* bit vector handling */
static elg_ui1 esd_count_bits           (elg_ui4              value);
static elg_ui4 esd_count_bits_in_buf    (const elg_ui4        grpc, 
                                         const elg_ui1*       grpv);
static UnifyGroupRep* esd_bitvec_encode (EsdUnifyData*        data,
                                         const elg_ui4        grpc, 
                                         const elg_ui1*       grpv);
static void esd_bitvec_decode           (EsdUnifyData*        data,
                                         elg_ui4              outc,
                                         elg_ui1*             outv,
                                         const UnifyGroupRep* grprep);
static void esd_bitvec_sparse_encode    (EsdUnifyData*        data,
                                         const elg_ui4        grpc, 
                                         const elg_ui1*       grpv,
                                         UnifyGroupRep*       grprep);
static void esd_bitvec_sparse_decode    (EsdUnifyData*        data,
                                         elg_ui4              grpc,
                                         elg_ui1*             grpv,
                                         const UnifyGroupRep* grprep);
static void esd_bitvec_inv_sparse_encode(EsdUnifyData*        data,
                                         const elg_ui4        grpc, 
                                         const elg_ui1*       grpv,
                                         UnifyGroupRep*       grprep);
static void esd_bitvec_inv_sparse_decode(EsdUnifyData*        data,
                                         elg_ui4              grpc,
                                         elg_ui1*             grpv,
                                         const UnifyGroupRep* grprep);
static void esd_bitvec_rle_encode       (EsdUnifyData*        data,
                                         const elg_ui4        grpc, 
                                         const elg_ui1*       grpv,
                                         UnifyGroupRep*       grprep);
static void esd_bitvec_rle_decode       (EsdUnifyData*        data,
                                         elg_ui4              grpc,
                                         elg_ui1*             grpv,
                                         const UnifyGroupRep* grprep);

/* File I/O */
static void write_map(const EsdUnifyData* instance, esd_unify_map_t map_id,
                      ElgOut* mapfile);


/*
 *---------------------------------------------------------------------------
 * EsdUnify
 *---------------------------------------------------------------------------
 */

/*--- Construction & destruction ------------------------------------------*/

/** 
 * @brief  Creates and returns an new instance of type EsdUnifyData.
 *
 * Creates and returns an new instance of type EsdUnifyData. If the
 * memory allocation request can not be fulfilled, an error message is
 * printed and the program is aborted.
 *
 * @return Pointer to new instance
 */
EsdUnifyData* esd_unify_create()
{
  EsdUnifyData* instance;

  /* Create data structure */
  instance = (EsdUnifyData*)malloc(sizeof(EsdUnifyData));
  if (!instance)
    elg_error();

  /* Initialize temporary data */
  instance->tmp_count  = 0;
  instance->tmp_string = NULL;
  instance->tmp_coord  = epk_vector_create_size(TMP_COORD_ARRAY);
  instance->grpcnt     = epk_vector_create_size(COMS_ARRAY);
  instance->comcnt     = epk_vector_create_size(WINDOWS_ARRAY);

  instance->tmp_bitvec      = NULL;
  instance->tmp_bitvec_size = 0;
  instance->tmp_max_procs   = 0;

  /* Initialize global data */
  instance->str2id  = epk_hashtab_create_size(STRINGS_HASH,
                                              (epk_ht_hash_f)&hash_string,
                                              (epk_ht_kcmp_f)&strcmp);
  instance->topmap  = epk_idmap_create(EPK_IDMAP_SPARSE, TOPOLOGIES_ARRAY);

  instance->strings = epk_vector_create_size(STRINGS_ARRAY);
  instance->machs   = epk_vector_create_size(MACHS_ARRAY);
  instance->nodes   = epk_vector_create_size(NODES_ARRAY);
  instance->procs   = epk_vector_create_size(PROCESS_ARRAY);
  instance->thrds   = epk_vector_create_size(THREADS_ARRAY);
  instance->locs    = epk_vector_create_size(LOCS_ARRAY);
  instance->files   = epk_vector_create_size(FILES_ARRAY);
  instance->regions = epk_vector_create_size(REGIONS_ARRAY);
  instance->csites  = epk_vector_create_size(CSITES_ARRAY);
  instance->cpaths  = epk_vector_create_size(CPATHS_ARRAY);
  instance->cproots = epk_vector_create();
  instance->metrics = epk_vector_create_size(METRICS_ARRAY);
  instance->coms    = epk_vector_create_size(COMS_ARRAY);
  instance->tops    = epk_vector_create_size(TOPOLOGIES_ARRAY);
  instance->wins    = epk_vector_create_size(WINDOWS_ARRAY);

  instance->mapoffs = epk_vector_create_size(OFFSET_ARRAY);

  /* Initialize per-process data */
  instance->maps[ESD_UNIFY_STRMAP  ] = epk_idmap_create(EPK_IDMAP_DENSE,  STRINGS_ARRAY);
  instance->maps[ESD_UNIFY_MACHMAP ] = epk_idmap_create(EPK_IDMAP_SPARSE, MACHS_ARRAY);
  instance->maps[ESD_UNIFY_NODEMAP ] = epk_idmap_create(EPK_IDMAP_SPARSE, NODES_ARRAY);
  instance->maps[ESD_UNIFY_LOCMAP  ] = epk_idmap_create(EPK_IDMAP_DENSE,  LOCS_ARRAY);
  instance->maps[ESD_UNIFY_FILEMAP ] = epk_idmap_create(EPK_IDMAP_DENSE,  FILES_ARRAY);
  instance->maps[ESD_UNIFY_REGMAP  ] = epk_idmap_create(EPK_IDMAP_DENSE,  REGIONS_ARRAY);
  instance->maps[ESD_UNIFY_CSITEMAP] = epk_idmap_create(EPK_IDMAP_DENSE,  CSITES_ARRAY);
  instance->maps[ESD_UNIFY_CPATHMAP] = epk_idmap_create(EPK_IDMAP_DENSE,  CPATHS_ARRAY);
  instance->maps[ESD_UNIFY_COMMAP  ] = epk_idmap_create(EPK_IDMAP_DENSE,  COMS_ARRAY);
  instance->maps[ESD_UNIFY_WINMAP  ] = epk_idmap_create(EPK_IDMAP_DENSE,  WINDOWS_ARRAY);

  instance->offcnt    = 0;
  instance->ltime[0]  = 0.0;
  instance->ltime[1]  = 1.0;
  instance->offset[0] = 0.0;
  instance->offset[1] = 0.0;

  instance->ntypes   = 0;
  instance->evtypes  = NULL;
  instance->evcounts = NULL;

  return instance;
}


/**
 * Destroys the given @a instance of EsdUnifyData and releases the allocated
 * memory.
 *
 * @param instance Object to be freed
 */
void esd_unify_free(EsdUnifyData* instance)
{
  int i;

  /* Validate arguments */
  assert(instance);

  /* Release temporary data */
  free(instance->tmp_string);
  epk_vector_foreach(instance->tmp_coord, (epk_vec_proc_f)&coord_free);
  epk_vector_free(instance->tmp_coord);
  epk_vector_foreach(instance->grpcnt, (epk_vec_proc_f)&grpcnt_free);
  epk_vector_free(instance->grpcnt);
  epk_vector_foreach(instance->comcnt, (epk_vec_proc_f)&free);
  epk_vector_free(instance->comcnt);
  free(instance->tmp_bitvec);

  /* Release global data */
  epk_hashtab_foreach(instance->str2id, &strhash_free);
  epk_hashtab_free(instance->str2id);
  epk_idmap_free(instance->topmap);

  epk_vector_free(instance->strings);
  epk_vector_foreach(instance->machs, &free);
  epk_vector_free(instance->machs);
  epk_vector_foreach(instance->nodes, &free);
  epk_vector_free(instance->nodes);
  epk_vector_foreach(instance->procs, &free);
  epk_vector_free(instance->procs);
  epk_vector_foreach(instance->thrds, &free);
  epk_vector_free(instance->thrds);
  epk_vector_foreach(instance->locs, &free);
  epk_vector_free(instance->locs);
  epk_vector_foreach(instance->files, &free);
  epk_vector_free(instance->files);
  epk_vector_foreach(instance->regions, &free);
  epk_vector_free(instance->regions);
  epk_vector_foreach(instance->csites, &free);
  epk_vector_free(instance->csites);
  epk_vector_foreach(instance->cpaths, (epk_vec_proc_f)&cpath_free);
  epk_vector_free(instance->cpaths);
  epk_vector_free(instance->cproots);
  epk_vector_foreach(instance->metrics, &free);
  epk_vector_free(instance->metrics);
  epk_vector_foreach(instance->coms, (epk_vec_proc_f)&com_free);
  epk_vector_free(instance->coms);
  epk_vector_foreach(instance->tops, (epk_vec_proc_f)&cart_free);
  epk_vector_free(instance->tops);
  epk_vector_foreach(instance->wins, &free);
  epk_vector_free(instance->wins);

  epk_vector_foreach(instance->mapoffs, &free);
  epk_vector_free(instance->mapoffs);

  /* Release per-process data */
  for (i = 0; i < ESD_UNIFY_NUM_MAPS; ++i)
    epk_idmap_free(instance->maps[i]);

  if (instance->evtypes)
    free(instance->evtypes);
  if (instance->evcounts)
    free(instance->evcounts);

  free(instance);
}


/*--- Unification ---------------------------------------------------------*/

/**
 * Processes @a size bytes of the local definition record data provided
 * in the given memory @a buffer. The unified definition data is collected
 * in the unification data object @a instance. In addition, the per-process
 * data (mapping tables etc.) is stored as well as written to the EPILOG
 * output stream @a mapfile (if defined).
 *
 * In case of an error (e.g., if the buffer does not contain EPILOG data),
 * an error message is printed and the program is aborted.
 *
 * @param instance Unification data object
 * @param buffer   Memory buffer containing the definition record data
 * @param size     Size of buffer in bytes
 * @param mapfile  EPILOG output stream
 */
void esd_unify_buffer(      EsdUnifyData* instance,
                      const buffer_t      buffer,
                            size_t        size,
                            ElgOut*       mapfile)
{
  ElgRCB*  handle;
  elg_ui4* offset;
  elg_ui4  rank;

  /* Validate arguments */
  assert(instance && buffer);

  /* Process buffer */
  handle = elg_read_open_buffer(buffer, size);
  if (!handle)
    elg_error();
  while (elg_read_next_def(handle, instance))
    ;
  elg_read_close(handle);

  if (!mapfile) return; /* skip file writing */

  /* Determine & store file offset in mapping file */
  rank   = epk_vector_size(instance->mapoffs);
  offset = (elg_ui4*)malloc(sizeof(elg_ui4));
  if (!offset)
    elg_error();
  *offset = ElgOut_tell(mapfile);
  epk_vector_push_back(instance->mapoffs, offset);

  /* Write section to mapping file */
  ElgOut_write_MAP_SECTION(mapfile, rank);

  if (instance->offcnt == 2) {
    ElgOut_write_OFFSET(mapfile, instance->ltime[0], instance->offset[0]);
    ElgOut_write_OFFSET(mapfile, instance->ltime[1], instance->offset[1]);
  } else if (instance->offcnt == 1) {
    elg_warning("Only one ELG_OFFSET record found -- discarded");
  }

  write_map(instance, ESD_UNIFY_LOCMAP,   mapfile);
  write_map(instance, ESD_UNIFY_REGMAP,   mapfile);
  write_map(instance, ESD_UNIFY_CSITEMAP, mapfile);
  write_map(instance, ESD_UNIFY_CPATHMAP, mapfile);
  write_map(instance, ESD_UNIFY_COMMAP,   mapfile);
  write_map(instance, ESD_UNIFY_WINMAP,   mapfile);

  if (instance->ntypes > 0) {
    ElgOut_write_EVENT_TYPES(mapfile, instance->ntypes, instance->evtypes);
    ElgOut_write_EVENT_COUNTS(mapfile, instance->ntypes, instance->evcounts);
  }
}


/*--- Mapping management --------------------------------------------------*/

/**
 * Returns one of the per-process identifier mapping tables stored in the
 * unification data object @a instance. The requested table is specified
 * by @a map_id.
 *
 * @param instance Unification data object
 * @param map_id   Requested mapping table
 *
 * @return Per-process identifier mapping table
 */
EpkIdMap* esd_unify_get_map(const EsdUnifyData*   instance,
                                  esd_unify_map_t map_id)
{
  /* Validate arguments */
  assert(ESD_UNIFY_NUM_MAPS > map_id);

  return instance->maps[map_id];
}


/**
 * Clears the per-process identifier mapping tables stored in the unification
 * data object @a instance.
 *
 * @param instance Unification data object
 *
 * @note This function does not modify/delete any global unification data.
 */
void esd_unify_clear_maps(EsdUnifyData* instance)
{
  int i;

  /* Validate arguments */
  assert(instance);

  /* Forget per-process data... */
  for (i = 0; i < ESD_UNIFY_NUM_MAPS; ++i)
    epk_idmap_clear(instance->maps[i]);

  instance->offcnt    = 0;
  instance->ltime[0]  = 0.0;
  instance->ltime[1]  = 1.0;
  instance->offset[0] = 0.0;
  instance->offset[1] = 0.0;

  free(instance->evtypes);
  free(instance->evcounts);

  instance->ntypes   = 0;
  instance->evtypes  = NULL;
  instance->evcounts = NULL;

  /* sequence counts are reset, bitvectors are retained */
  epk_vector_foreach(instance->grpcnt, (epk_vec_proc_f)&grpcnt_reset);

  epk_vector_foreach(instance->comcnt, &free);
  epk_vector_clear(instance->comcnt);
}


/*--- File I/O ------------------------------------------------------------*/

/**
 * Writes the unified definition data collected in the given EsdUnifyData
 * @a instance to the EPILOG output stream @a deffile.
 *
 * @param instance Object containing the unified definition data
 * @param deffile  EPILOG output stream
 */
void esd_unify_write_defs(EsdUnifyData* instance, ElgOut* deffile)
{
  void** ptr;
  void** end;
  size_t index;

  /* Validate arguments */
  assert(instance && deffile);

  /* Write string definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->strings);
  end   = epk_vector_end(instance->strings);
  while (ptr != end) {
    char** entry = (char**)ptr;
    ElgOut_write_string(deffile, index++, *entry);
    ptr++;
  }

  /* Write machine definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->machs);
  end   = epk_vector_end(instance->machs);
  while (ptr != end) {
    UnifyMach* entry = (UnifyMach*)(*ptr);
    ElgOut_write_MACHINE(deffile, index++, entry->nodec, entry->nameid);
    ptr++;
  }

  /* Write node definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->nodes);
  end   = epk_vector_end(instance->nodes);
  while (ptr != end) {
    UnifyNode* entry = (UnifyNode*)(*ptr);
    ElgOut_write_NODE(deffile, index++, entry->machid, entry->cpuc,
                      entry->nameid, entry->clockrt);
    ptr++;
  }

  /* Write process definitions */
  ptr   = epk_vector_begin(instance->procs);
  end   = epk_vector_end(instance->procs);
  while (ptr != end) {
    UnifyProc* entry = (UnifyProc*)(*ptr);
    ElgOut_write_PROCESS(deffile, entry->id, entry->nameid);
    ptr++;
  }

  /* Write thread definitions */
  ptr   = epk_vector_begin(instance->thrds);
  end   = epk_vector_end(instance->thrds);
  while (ptr != end) {
    UnifyThrd* entry = (UnifyThrd*)(*ptr);
    ElgOut_write_THREAD(deffile, entry->id, entry->pid, entry->nameid);
    ptr++;
  }

  /* Write location definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->locs);
  end   = epk_vector_end(instance->locs);
  while (ptr != end) {
    UnifyLoc* entry = (UnifyLoc*)(*ptr);
    ElgOut_write_LOCATION(deffile, index++, entry->machid, entry->nodeid,
                          entry->procid, entry->thrdid);
    ptr++;
  }

  /* Write file definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->files);
  end   = epk_vector_end(instance->files);
  while (ptr != end) {
    elg_ui4* entry = (elg_ui4*)(*ptr);
    ElgOut_write_FILE(deffile, index++, *entry);
    ptr++;
  }

  /* Write region definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->regions);
  end   = epk_vector_end(instance->regions);
  while (ptr != end) {
    UnifyRegion* entry = (UnifyRegion*)(*ptr);
    ElgOut_write_REGION(deffile, index++, entry->nameid, entry->fileid,
                        entry->begln, entry->endln, entry->descid, entry->type);
    ptr++;
  }

  /* Write call site definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->csites);
  end   = epk_vector_end(instance->csites);
  while (ptr != end) {
    UnifyCsite* entry = (UnifyCsite*)(*ptr);
    ElgOut_write_CALL_SITE(deffile, index++, entry->fileid, entry->lineno,
                           entry->e_regid, entry->l_regid);
    ptr++;
  }

  /* Write call path definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->cpaths);
  end   = epk_vector_end(instance->cpaths);
  while (ptr != end) {
    UnifyCpath* entry = (UnifyCpath*)(*ptr);
    ElgOut_write_CALL_PATH(deffile, index++, entry->rid, entry->ppid,
                           entry->order);
    ptr++;
  }

  /* Write metric definitions */
  ptr   = epk_vector_begin(instance->metrics);
  end   = epk_vector_end(instance->metrics);
  while (ptr != end) {
    UnifyMetric* entry = (UnifyMetric*)(*ptr);
    ElgOut_write_METRIC(deffile, entry->id, entry->nameid, entry->descid,
                        entry->type, entry->mode, entry->ival);
    ptr++;
  }

  /* Write communicator definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->coms);
  end   = epk_vector_end(instance->coms);
  while (ptr != end) {
    UnifyCom* entry = (UnifyCom*)(*ptr);
    esd_bitvec_decode(instance, instance->tmp_bitvec_size, 
                      instance->tmp_bitvec, entry->group->vec);
    ElgOut_write_MPI_COMM(deffile, index++, entry->uag, 
                          instance->tmp_bitvec_size, instance->tmp_bitvec);
    ptr++;
  }

  /* Write topology definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->tops);
  end   = epk_vector_end(instance->tops);
  while (ptr != end) {
    void**     coord_ptr;
    void**     coord_end;
    UnifyCart* entry = (UnifyCart*)(*ptr);
    ElgOut_write_CART_TOPOLOGY(deffile, index, entry->comid, entry->ndims,
                               entry->nlocs, entry->period);

    coord_ptr = epk_vector_begin(entry->coords);
    coord_end = epk_vector_end(entry->coords);
    while (coord_ptr != coord_end) {
      CartCoord* centry = (CartCoord*)(*coord_ptr);
      ElgOut_write_CART_COORDS(deffile, index, centry->locid, entry->ndims,
                               centry->coord);
      coord_ptr++;
    }

    index++;
    ptr++;
  }

  /* Write RMA window definitions */
  index = 0;
  ptr   = epk_vector_begin(instance->wins);
  end   = epk_vector_end(instance->wins);
  while (ptr != end) {
    UnifyWin* entry = (UnifyWin*)(*ptr);
    ElgOut_write_MPI_WIN(deffile, index++, entry->comid);
    ptr++;
  }

  /* Write mapping file offsets */
  index = 0;
  ptr   = epk_vector_begin(instance->mapoffs);
  end   = epk_vector_end(instance->mapoffs);
  while (ptr != end) {
    elg_ui4* offset = (elg_ui4*)(*ptr);
    ElgOut_write_MAP_OFFSET(deffile, index++, *offset);
    ptr++;
  }
}


/*
 *---------------------------------------------------------------------------
 * Local functions
 *---------------------------------------------------------------------------
 */

/*--- Hash functions ------------------------------------------------------*/

/*
 * Returns a hash value for the given string @a key.
 */
size_t hash_string(const char* key)
{
  size_t hash;

  /* Validate arguments */
  assert(key);

  /* Calculate hash value */
  hash = 5381;
  while (*key) {
    hash = hash * 33 + *key;
    key++;
  }

  return hash;
}


/*--- Unification functions -----------------------------------------------*/

/*
 * Helper function for the unification of strings. It searches for an existing
 * definition of the string @a key in the unfication data object @a data and
 * eventually adds it. In addition, an appropriate mapping from the given
 * @a local_id to its global identifier is created.
 */
void unify_string(EsdUnifyData* data, char* key, elg_ui4 local_id)
{
  EpkHashEntry* entry;
  size_t        index;
  elg_ui4*      global_id;

  /* String not yet defined? */
  entry = epk_hashtab_find(data->str2id, key, &index);
  if (!entry) {
    /* Create new entry */
    global_id = (elg_ui4*)malloc(sizeof(elg_ui4));
    if (!global_id)
      elg_error();

    /* Determine global ID */
    *global_id = epk_vector_size(data->strings);
    
    /* Insert string into tables */
    epk_hashtab_insert(data->str2id, key, global_id, &index);
    epk_vector_push_back(data->strings, key);
  } else {
    /* Retrieve global ID */
    global_id = (elg_ui4*)entry->value;

    /* Release temporary string */
    free(key);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_STRMAP], local_id, *global_id);
}


/*--- Comparisons ---------------------------------------------------------*/

/*
 * Compares two instances of type UnifySingleId (or a derived type) for
 * equality by comparing the IDs. Returns 0 if the IDs are equal and 1
 * otherwise.
 */
int single_id_cmp(const UnifySingleId* key, const UnifySingleId* entry)
{
  return !(key->id == entry->id);
}


/*
 * Compares two instances of type UnifyDoubleId (or a derived type) for
 * equality by comparing the IDs. Returns 0 if both IDs are equal and
 * 1 otherwise.
 */
int double_id_cmp(const UnifyDoubleId* key, const UnifyDoubleId* entry)
{
  return !(key->id1 == entry->id1 &&
           key->id2 == entry->id2);
}


/*
 * Compares the two given region definitions for equality. Returns 0 if
 * they are equal and 1 otherwise.
 */
int region_cmp(const UnifyRegion* key, const UnifyRegion* entry)
{
  return !(key->nameid == entry->nameid &&
           key->fileid == entry->fileid &&
           key->begln  == entry->begln  &&
           key->endln  == entry->endln);
}


/*
 * Compares the two given call site definitions for equality. Returns 0 if
 * they are equal and 1 otherwise.
 */
int csite_cmp(const UnifyCsite* key, const UnifyCsite* entry)
{
  return !(key->fileid  == entry->fileid  &&
           key->lineno  == entry->lineno  &&
           key->e_regid == entry->e_regid &&
           key->l_regid == entry->l_regid);
}


/*
 * Compares the two given MPI communicator definitions for equality. Returns
 * 0 if they are equal and 1 otherwise.
 */
int com_cmp(const UnifyCom* key, const UnifyCom* entry)
{
  /* Compare group pointer & sequence number */
  if (key->group != entry->group || key->seqno != entry->seqno)
    return 1;

  return 0;
}


/*
 * Compares the two given process group definitions for equality. Returns
 * 0 if they are equal and 1 otherwise.
 */
int grpcnt_cmp(const GrpCount* key, const GrpCount* entry)
{
  assert(key);
  assert(key->group);
  assert(key->group->vec);

  assert(entry);
  assert(entry->group);
  assert(entry->group->vec);

  /* Compare length */
  if (key->group->vec->size != entry->group->vec->size)
    return 1;

  /* Compare bitstring */
  if ((key->group->vec->buf == NULL) && (entry->group->vec->buf == NULL))
    return 0;
  else
    return memcmp(key->group->vec->buf, entry->group->vec->buf, 
                  key->group->vec->size);
}


/*--- Memory management ---------------------------------------------------*/

/*
 * Releases the memory occupied by the given string hash table entry.
 */
void strhash_free(EpkHashEntry* entry)
{
  free(entry->key);
  free(entry->value);
}


/*
 * Releases the memory occupied by the given call path definition entry.
 */
void cpath_free(UnifyCpath* entry)
{
  if (entry->children)
    epk_vector_free(entry->children);
  free(entry);
}


/*
 * Releases the memory occupied by the given MPI communicator definition
 * entry.
 */
void com_free(UnifyCom* entry)
{
  /* decrease reference count on group*/
  entry->group->refs--;
  
  /* check for deletion */
  if (entry->group->refs == 0)
  {
    free(entry->group->vec->buf);
    free(entry->group->vec);
    free(entry->group); 
  }
  free(entry);
}


/*
 * Releases the memory occupied by the given cartesian topology definition
 * entry.
 */
void cart_free(UnifyCart* entry)
{
  /* Release coordinates */
  epk_vector_foreach(entry->coords, (epk_vec_proc_f)&coord_free);
  epk_vector_free(entry->coords);

  /* Release data structure */
  free(entry->nlocs);
  free(entry->period);
  free(entry);
}


/*
 * Releases the memory occupied by the given cartesian coordinate definition
 * entry.
 */
void coord_free(CartCoord* entry)
{
  free(entry->coord);
  free(entry);
}

/*
 * Resets the group counter to start with fresh counts on a new location
 * while retaining the bitvectors themselves.
 */
void grpcnt_reset(GrpCount* entry)
{
  assert(entry);

  entry->count = 0;
}

/*
 * Releases the memory occupied by the given group counter entry.
 */
void grpcnt_free(GrpCount* entry)
{
  assert(entry);
  assert(entry->group);
  assert(entry->group->vec);

  /* decrease reference count on group*/
  entry->group->refs--;
  
  /* check for deletion */
  if (entry->group->refs == 0)
  {
    free(entry->group->vec->buf);
    free(entry->group->vec);
    free(entry->group); 
  }
  free(entry);
}


/*--- File I/O ------------------------------------------------------------*/

/*
 * Writes the mapping stored in the identifier mapping table with ID
 * @a map_id of the unification data object @a instance to the EPILOG output
 * stream @a mapfile.
 */
void write_map(const EsdUnifyData* instance, esd_unify_map_t map_id,
               ElgOut* mapfile)
{
  EpkIdMap* map;
  elg_ui1   type;
  elg_ui1   mode;
  size_t    count;
  elg_ui4*  data;

  /* Determine map */
  map = esd_unify_get_map(instance, map_id);

  /* Convert mapping table ID */
  switch (map_id) {
    case ESD_UNIFY_LOCMAP:
      type = ELG_LOCATION;
      break;
    case ESD_UNIFY_REGMAP:
      type = ELG_REGION;
      break;
    case ESD_UNIFY_CSITEMAP:
      type = ELG_CALL_SITE;
      break;
    case ESD_UNIFY_CPATHMAP:
      type = ELG_CALL_PATH;
      break;
    case ESD_UNIFY_COMMAP:
      type = ELG_MPI_COMM;
      break;
    case ESD_UNIFY_WINMAP:
      type = ELG_MPI_WIN;
      break;
    default:
      elg_error_msg("Unknown mapping table ID");
  }

  /* Convert mapping mode */
  mode = epk_idmap_mode(map) == EPK_IDMAP_DENSE ? ELG_MAP_DENSE
                                                : ELG_MAP_SPARSE;

  /* Query data */
  data = (elg_ui4*)epk_idmap_data(map, &count);

  /* Write record */
  if (count > 0)
    ElgOut_write_IDMAP(mapfile, type, mode, count, data);
}


/*
 *---------------------------------------------------------------------------
 * EPILOG read callbacks
 *---------------------------------------------------------------------------
 */

void elg_readcb_ALL(elg_ui1 type,         /* record type */
		    elg_ui1 length,       /* record length */
                    void*   userdata)
{
  /* Nothing to be done... */
}


void elg_readcb_STRING(elg_ui4     strid,        /* string identifier */
		       elg_ui1     cntc,         /* number of continuation records */
		       const char* str,          /* string */
                       void*       userdata) 
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;

  /* Safety check */
  if (data->tmp_count > 0)
    elg_error_msg("ELG_STRING_CNT record expected!");

  /* No continuation records? */
  if (cntc == 0) {
    char* key;

    /* Duplicate string */
    key = strdup(str);
    if (!key)
      elg_error();

    /* Try to unify string */
    unify_string(data, key, strid);  
  } else {
    elg_ui1 maxlen;

    /* Save temporary data */
    data->tmp_id     = strid;
    data->tmp_count  = cntc;
    data->tmp_string = (char*)malloc((cntc + 1) * 255 * sizeof(char));
    if (!data->tmp_string)
      elg_error();

    maxlen = 255 - sizeof(elg_ui4) - sizeof(elg_ui1);
    strncpy(data->tmp_string, str, maxlen);
    data->tmp_string[maxlen] = '\0';
  }
}


void elg_readcb_STRING_CNT(const char* str,          /* continued string */
                           void*       userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;

  /* Safety check */
  if (data->tmp_count == 0)
    elg_error_msg("Unexpected ELG_STRING_CNT record!");

  /* Concatenate substrings */
  strncat(data->tmp_string, str, 255);

  /* Eventually try to unify string */
  data->tmp_count--;
  if (data->tmp_count == 0) {
    unify_string(data, data->tmp_string, data->tmp_id);
    data->tmp_string = NULL;
  }
}


void elg_readcb_MACHINE(elg_ui4 mid,          /* machine identifier */
		        elg_ui4 nodec,        /* number of nodes */
		        elg_ui4 mnid,         /* machine name identifier */
                        void*   userdata) 
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  size_t        global_id;

  /* Machine not yet defined? */
  if (epk_vector_find(data->machs, &mid,
                      (epk_vec_cmp_f)&single_id_cmp, &global_id) == 0) {
    UnifyMach* entry;

    /* Determine global ID */
    global_id = epk_vector_size(data->machs);

    /* Create new machine definition record */
    entry = (UnifyMach*)malloc(sizeof(UnifyMach));
    if (!entry)
      elg_error();
    entry->id     = mid;
    entry->nodec  = 0;
    entry->nameid = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], mnid);

    epk_vector_push_back(data->machs, entry);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_MACHMAP], mid, global_id);
}


void elg_readcb_NODE(elg_ui4 nid,          /* node identifier */
		     elg_ui4 mid,          /* machine identifier */
		     elg_ui4 cpuc,         /* number of CPUs */
		     elg_ui4 nnid,         /* node name identifer */
		     elg_d8  cr,           /* number of clock cycles/sec */
                     void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  UnifyNode*    entry;
  size_t        global_id;

  /* Construct new entry */
  entry = (UnifyNode*)malloc(sizeof(UnifyNode));
  if (!entry)
      elg_error();
  entry->id      = nid;
  entry->machid  = epk_idmap_get(data->maps[ESD_UNIFY_MACHMAP], mid);
  entry->cpuc    = cpuc;
  entry->nameid  = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], nnid);
  entry->clockrt = cr;

  /* Node not yet defined? */
  if (epk_vector_find(data->nodes, entry,
                      (epk_vec_cmp_f)&double_id_cmp, &global_id) == 0) {
    /* Determine global ID */
    global_id = epk_vector_size(data->nodes);

    /* Adjust node count for machine */
    ((UnifyMach*)epk_vector_at(data->machs, entry->machid))->nodec++;

    epk_vector_push_back(data->nodes, entry);
  } else {
    free(entry);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_NODEMAP], nid, global_id);
}


void elg_readcb_PROCESS(elg_ui4 pid,          /* process identifier */
		        elg_ui4 pnid,         /* process name identifier */
                        void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;

  /* Process not yet defined? */
  if (epk_vector_find(data->procs, &pid,
                      (epk_vec_cmp_f)&single_id_cmp, NULL) == 0) {
    UnifyProc* entry;

    /* Create new process definition record */
    entry = (UnifyProc*)malloc(sizeof(UnifyProc));
    if (!entry)
      elg_error();
    entry->id     = pid;
    entry->nameid = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], pnid);

    epk_vector_push_back(data->procs, entry);
  }
}


void elg_readcb_THREAD(elg_ui4 tid,          /* thread identifier */
		       elg_ui4 pid,          /* process identifier */
		       elg_ui4 tnid,         /* thread name identifier */
                       void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  UnifyThrd*    entry;

  /* Create new thread definition record */
  entry = (UnifyThrd*)malloc(sizeof(UnifyThrd));
  if (!entry)
    elg_error();
  entry->id     = tid;
  entry->pid    = pid;
  entry->nameid = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], tnid);

  /* Thread not yet defined? */
  if (epk_vector_find(data->thrds, entry,
                      (epk_vec_cmp_f)&double_id_cmp, NULL) == 0)
    epk_vector_push_back(data->thrds, entry);
  else
    free(entry);
}


void elg_readcb_LOCATION(elg_ui4 lid,          /* location identifier */
			 elg_ui4 mid,          /* machine identifier */
			 elg_ui4 nid,          /* node identifier */
			 elg_ui4 pid,          /* process identifier */
			 elg_ui4 tid,          /* thread identifier */
                         void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  UnifyLoc*     entry;
  elg_ui4       global_id;
  size_t        index;

  /*
   * It is assumed that each location is defined only once, so it
   * is not necessary to check whether it is already defined.  
   */

  /* Determine global ID */
  global_id = epk_vector_size(data->locs);

  /* Create new location definition record */
  entry = (UnifyLoc*)malloc(sizeof(UnifyLoc));
  if (!entry)
    elg_error();

  entry->machid = epk_idmap_get(data->maps[ESD_UNIFY_MACHMAP], mid);
  entry->nodeid = epk_idmap_get(data->maps[ESD_UNIFY_NODEMAP], nid);
  entry->procid = pid;
  entry->thrdid = tid;

  epk_vector_push_back(data->locs, entry);

  /* Add mappings */
  epk_idmap_add(data->maps[ESD_UNIFY_LOCMAP], lid, global_id);

  /* Process temporary cartesian coordinate buffer */
  index = epk_vector_size(data->tmp_coord);
  while (index > 0) {
    CartCoord* entry = (CartCoord*)epk_vector_at(data->tmp_coord, index - 1);

    /* Eventually store global location ID & move entry to topology */
    if (entry->locid == lid) {
      UnifyCart* topology = (UnifyCart*)epk_vector_at(data->tops, entry->topid);

      entry->locid = global_id;
      epk_vector_push_back(topology->coords, entry);
      epk_vector_erase(data->tmp_coord, index - 1);
    }

    index--;
  }
}


void elg_readcb_FILE(elg_ui4 fid,          /* file identifier */
		     elg_ui4 fnid,         /* file name identifier */
                     void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  elg_ui4       name;
  size_t        global_id;

  /* Map file name */
  name = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], fnid);

  /* File not yet defined? */
  if (epk_vector_find(data->files, &name,
                      (epk_vec_cmp_f)&single_id_cmp, &global_id) == 0) {
    elg_ui4* entry;

    /* Determine global ID */
    global_id = epk_vector_size(data->files);

    /* Create new entry */
    entry = (elg_ui4*)malloc(sizeof(elg_ui4));
    if (!entry)
      elg_error();
    *entry = name;

    epk_vector_push_back(data->files, entry);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_FILEMAP], fid, global_id);
}


void elg_readcb_REGION(elg_ui4 rid,          /* region identifier */
		       elg_ui4 rnid,         /* region name identifier */
		       elg_ui4 fid,          /* source file identifier */
		       elg_ui4 begln,        /* begin line number */
		       elg_ui4 endln,        /* end line number */
		       elg_ui4 rdid,         /* region description identifier */
		       elg_ui1 rtype,        /* region type */
                       void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  UnifyRegion*  entry;
  size_t        global_id;

  /* Construct new entry */
  entry = (UnifyRegion*)malloc(sizeof(UnifyRegion));
  if (!entry)
      elg_error();
  entry->nameid = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], rnid);
  entry->fileid = epk_idmap_get(data->maps[ESD_UNIFY_FILEMAP], fid);
  entry->begln  = begln;
  entry->endln  = endln;
  entry->descid = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], rdid);
  entry->type   = rtype;

  /* Region not yet defined? */
  if (epk_vector_find(data->regions, entry,
                      (epk_vec_cmp_f)&region_cmp, &global_id) == 0) {
    /* Determine global ID */
    global_id = epk_vector_size(data->regions);

    epk_vector_push_back(data->regions, entry);
  } else {
    free(entry);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_REGMAP], rid, global_id);
}


void elg_readcb_CALL_SITE(elg_ui4 csid,         /* call site identifier */
			  elg_ui4 fid,          /* source file identifier */
			  elg_ui4 lno,          /* line number */
			  elg_ui4 erid,         /* region identifer to be entered */
			  elg_ui4 lrid,         /* region identifer to be left */
                          void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  UnifyCsite*   entry;
  size_t        global_id;

  /* Construct new entry */
  entry = (UnifyCsite*)malloc(sizeof(UnifyCsite));
  if (!entry)
      elg_error();
  entry->fileid  = epk_idmap_get(data->maps[ESD_UNIFY_FILEMAP], fid);
  entry->lineno  = lno;
  entry->e_regid = epk_idmap_get(data->maps[ESD_UNIFY_REGMAP], erid);
  entry->l_regid = epk_idmap_get(data->maps[ESD_UNIFY_REGMAP], lrid);

  /* Call site not yet defined? */
  if (epk_vector_find(data->csites, entry,
                      (epk_vec_cmp_f)&csite_cmp, &global_id) == 0) {
    /* Determine global ID */
    global_id = epk_vector_size(data->csites);

    epk_vector_push_back(data->csites, entry);
  } else {
    free(entry);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_CSITEMAP], csid, global_id);
}


void elg_readcb_CALL_PATH(elg_ui4 cpid,         /* call-path identifier */
			  elg_ui4 rid,          /* node region identifier */
			  elg_ui4 ppid,         /* parent call-path identifier */
			  elg_ui8 order,        /* node order specifier */
                          void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  EpkVector*    list;
  UnifyCpath*   entry;
  size_t        index;
  size_t        global_id;

  /* Construct new entry */
  entry = (UnifyCpath*)malloc(sizeof(UnifyCpath));
  if (!entry)
    elg_error();
  entry->rid  = epk_idmap_get(data->maps[ESD_UNIFY_REGMAP], rid);
  entry->ppid = epk_idmap_get(data->maps[ESD_UNIFY_CPATHMAP], ppid);

  /* Root call-path? => Search in root list */
  /* Otherwise search in parents child list */
  if (ELG_NO_ID == entry->ppid)
    list = data->cproots;
  else {
    UnifyCpath* parent = (UnifyCpath*)epk_vector_at(data->cpaths, entry->ppid);

    /* Eventually create child list */
    if (!parent->children)
      parent->children = epk_vector_create();

    list = parent->children;
  }

  /* Call-path not yet defined? */
  if (epk_vector_find(list, entry,
                      (epk_vec_cmp_f)&double_id_cmp, &index) == 0) {
    /* Determine global ID */
    global_id = epk_vector_size(data->cpaths);

    /* Initialize remaining fields */
    entry->id       = global_id;
    entry->order    = order;
    entry->children = NULL;

    /* Add item to call-path vector & root/child list */
    epk_vector_push_back(data->cpaths, entry);
    epk_vector_push_back(list, entry);
  } else {
    UnifyCpath* exist = epk_vector_at(list, index);

    /* Determine global ID & update order attribute */
    global_id = exist->id;
    if (exist->order > order)
      exist->order = order;

    free(entry);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_CPATHMAP], cpid, global_id);
}


void elg_readcb_METRIC(elg_ui4 metid,        /* metric identifier */
		       elg_ui4 metnid,       /* metric name identifier */
		       elg_ui4 metdid,       /* metric description identifier */
		       elg_ui1 metdtype,     /* metric data type */
		       elg_ui1 metmode,      /* metric mode */
		       elg_ui1 metiv,        /* time interval referenced */
                       void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  UnifyMetric*  entry;

  /* 
   * It is assumed that each metric carries its global
   * identifier.
   *
   * Also, it is assumed that the metric identifiers are
   * already enumerated from 0 to n-1.
   */

  /* Create new metric definition record */
  entry = (UnifyMetric*)malloc(sizeof(UnifyMetric));
  if (!entry)
    elg_error();
  entry->id     = metid;
  entry->nameid = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], metnid);
  entry->descid = epk_idmap_get(data->maps[ESD_UNIFY_STRMAP], metdid);
  entry->type   = metdtype;
  entry->mode   = metmode;
  entry->ival   = metiv;

  /* Metric not yet defined? */
  if (epk_vector_find(data->metrics, entry,
                      (epk_vec_cmp_f)&single_id_cmp, NULL) == 0)
    epk_vector_push_back(data->metrics, entry);
  else
    free(entry);
}


void elg_readcb_MPI_COMM(elg_ui4 cid,          /* communicator identifier */
                         elg_ui1 uag,          /* used as group */
                         elg_ui4 grpc,         /* size if bitstring in bytes */
                         elg_ui1 grpv[],       /* bitstring defining the group */
                         void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  GrpCount*      counter = NULL;
  UnifyGroup*    ingroup = NULL;
  UnifyCom*      entry   = NULL;
  size_t         index;
  size_t         global_id;

  assert(data);

  /* initialize temporary buffer */
  if (!data->tmp_bitvec)
  {
    elg_cntl_msg("Initialising temporary buffer for group vector compression.");
    data->tmp_bitvec_size = grpc * sizeof(elg_ui1);
    data->tmp_bitvec      = (elg_ui1*) malloc(data->tmp_bitvec_size);
    if (!data->tmp_bitvec)
      elg_error();
  }

  /* initialize UnifyGroup structure */
  ingroup = (UnifyGroup*)malloc(sizeof(UnifyGroup));
  if (!ingroup)
    elg_error();
  ingroup->refs = 0;
  ingroup->vec  = esd_bitvec_encode(data, grpc, grpv);

  /* Create counter entry */
  counter = (GrpCount*)malloc(sizeof(GrpCount));
  if (!counter)
    elg_error();
  /*
   * Groups will always have sequence number 0
   * Communicators have sequence numbers starting at 1
   */
  counter->count = 1;
  counter->group = ingroup;
  counter->group->refs++;

  /* Search counter entry */
  if (epk_vector_find(data->grpcnt, counter,
                      (epk_vec_cmp_f)&grpcnt_cmp, &index) == 0) {
    epk_vector_push_back(data->grpcnt, counter);
  } else {
    grpcnt_free(counter);
    counter = (GrpCount*)epk_vector_at(data->grpcnt, index);

    if (uag == 0)
    {
      counter->count++;
    }
  }

  /* Create communicator description record */
  entry = (UnifyCom*)malloc(sizeof(UnifyCom));
  if (!entry)
    elg_error();
  entry->seqno = uag ? 0 : counter->count;
  entry->uag   = uag;
  entry->group = counter->group;
  /* increment reference count on group */
  entry->group->refs++;

  /* Communicator not yet defined? */
  if (epk_vector_find(data->coms, entry,
                      (epk_vec_cmp_f)&com_cmp, &global_id) == 0) {
    /* Determine global ID */
    global_id = epk_vector_size(data->coms);

    epk_vector_push_back(data->coms, entry);
  } else {
    com_free(entry);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_COMMAP], cid, global_id);
}


void elg_readcb_CART_TOPOLOGY(elg_ui4 topid,         /* topology id */
                              elg_ui4 cid,           /* communicator id */
			      elg_ui1 ndims,         /* number of dimensions */
			      elg_ui4 dimv[],        /* number of processes in each dim */  
			      elg_ui1 periodv[],     /* periodicity in each dim */
			      void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  elg_ui4       gcid = epk_idmap_get(data->maps[ESD_UNIFY_COMMAP], cid);
  size_t        index;

  /* Topology not yet defined? */
  if (epk_vector_find(data->tops, &gcid,
                      (epk_vec_cmp_f)&single_id_cmp, &index) == 0) {
    elg_ui4    ncoords;
    elg_ui4    i;
    UnifyCart* entry;

    /* Calculate total number of coordinates */
    ncoords = 1;
    for (i = 0; i < ndims; ++i)
      ncoords *= dimv[i];

    /* Create new topology entry */
    entry = (UnifyCart*)malloc(sizeof(UnifyCart));
    if (!entry)
      elg_error();
    entry->comid  = gcid;
    entry->ndims  = ndims;
    entry->nlocs  = (elg_ui4*)malloc(ndims * sizeof(elg_ui4));
    if (!entry->nlocs)
      elg_error();
    memcpy(entry->nlocs, dimv, ndims * sizeof(elg_ui4));
    entry->period = (elg_ui1*)malloc(ndims * sizeof(elg_ui1));
    if (!entry->period)
      elg_error();
    memcpy(entry->period, periodv, ndims * sizeof(elg_ui1));
    entry->coords = epk_vector_create_size(ncoords);

    /* Store new topology */
    epk_idmap_add(data->topmap, gcid, epk_vector_size(data->tops));
    epk_vector_push_back(data->tops, entry);
  }
}


void elg_readcb_CART_COORDS(elg_ui4 topid,        /* topology id */
			    elg_ui4 lid,          /* location id */
			    elg_ui1 ndims,        /* number of dimensions */
			    elg_ui4 coordv[],     /* coordinates in each dimension */
			    void*   userdata)
{
  EsdUnifyData* data   = (EsdUnifyData*)userdata;
  elg_ui4       gtopid = epk_idmap_get(data->maps[ESD_UNIFY_COMMAP], topid);
  CartCoord*    entry;

  /* Search topology (topology ID == communicator ID) */
  if (epk_vector_find(data->tops, &gtopid,
                      (epk_vec_cmp_f)&single_id_cmp, NULL) == 0)
    elg_error_msg("ELG_CART_COORDS record refers to undefined topology!");

  /* Create new coordinate entry */
  entry = (CartCoord*)malloc(sizeof(CartCoord));
  if (!entry)
    elg_error();
  entry->locid = lid;
  entry->topid = epk_idmap_get(data->topmap, gtopid);
  entry->coord = (elg_ui4*)malloc(ndims * sizeof(elg_ui4));
  if (!entry->coord)
    elg_error();
  memcpy(entry->coord, coordv, ndims * sizeof(elg_ui4));

  /* Temporarily buffer coordinate entry, it will */
  /* be processed when location record is found */
  epk_vector_push_back(data->tmp_coord, entry);
}


void elg_readcb_OFFSET(elg_d8 ltime,        /* local time */
		       elg_d8 offset,       /* offset to global time */
                       void*  userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;

  /* Sanity check */
  if (data->offcnt == 2)
    elg_error_msg("Unexpected ELG_OFFSET record!");

  /* Store offset data */
  data->ltime [data->offcnt] = ltime;
  data->offset[data->offcnt] = offset;
  data->offcnt++;
}


void elg_readcb_NUM_EVENTS(elg_ui4 eventc,       /* number of event records */
                           void*   userdata)
{
  /* Nothing to be done... */
}


void elg_readcb_EVENT_TYPES(elg_ui4 ntypes,     /* number of event types */
                            elg_ui1 typev[],    /* vector of event types */
                            void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;

  /* Sanity checks */
  if (data->ntypes > 0 && data->ntypes != ntypes)
    elg_error_msg("Inconsistent ELG_EVENT_TYPES record!");
  if (data->evtypes)
    elg_error_msg("Unexpected ELG_EVENT_TYPES record!");

  /* Store type data */
  data->ntypes  = ntypes;
  data->evtypes = (elg_ui1*)malloc(ntypes * sizeof(elg_ui1));
  memcpy(data->evtypes, typev, ntypes * sizeof(elg_ui1));
}


void elg_readcb_EVENT_COUNTS(elg_ui4 ntypes,    /* number of event types */
                             elg_ui4 countv[],  /* vector of type counts */
                             void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;

  /* Sanity checks */
  if (data->ntypes > 0 && data->ntypes != ntypes)
    elg_error_msg("Inconsistent ELG_EVENT_COUNTS record!");
  if (data->evcounts)
    elg_error_msg("Unexpected ELG_EVENT_COUNTS record!");

  /* Store type data */
  data->ntypes   = ntypes;
  data->evcounts = (elg_ui4*)malloc(ntypes * sizeof(elg_ui4));
  memcpy(data->evcounts, countv, ntypes * sizeof(elg_ui4));
}


void elg_readcb_MAP_SECTION(elg_ui4 rank,   /* rank number the section relates to */
                            void* userdata)
{
  /* Nothing to be done... */
}


void elg_readcb_MAP_OFFSET(elg_ui4 rank,   /* rank number the offset relates to */
                           elg_ui4 offset, /* file offset in mapping file */
                           void* userdata )
{
  /* Nothing to be done... */
}

void elg_readcb_IDMAP(elg_ui1 type,     /* object type to be mapped */
                      elg_ui1 mode,     /* mapping mode (dense/sparse) */
                      elg_ui4 count,    /* number of entries */
                      elg_ui4 mapv[],   /* vector of mappings */
                      void* userdata)
{
  /* Nothing to be done... */
}


void elg_readcb_LAST_DEF(void* userdata)
{
  /* Nothing to be done... */
}


void elg_readcb_MPI_WIN(elg_ui4 wid,          /* window identifier */
                        elg_ui4 cid,          /* communicator identifier */
                        void*   userdata)
{
  EsdUnifyData* data = (EsdUnifyData*)userdata;
  ComCount*     counter;
  UnifyWin*     entry;
  size_t        index;
  size_t        global_id;

  /* Create counter entry */
  counter = (ComCount*)malloc(sizeof(ComCount));
  if (!counter)
    elg_error();
  counter->comid = epk_idmap_get(data->maps[ESD_UNIFY_COMMAP], cid);
  counter->count = 0;

  /* Search counter entry */
  if (epk_vector_find(data->comcnt, counter,
                      (epk_vec_cmp_f)&single_id_cmp, &index) == 0) {
    epk_vector_push_back(data->comcnt, counter);
  } else {
    free(counter);
    counter = (ComCount*)epk_vector_at(data->comcnt, index);
    counter->count++;
  }

  /* Create RMA window description record */
  entry = (UnifyWin*)malloc(sizeof(UnifyWin));
  if (!entry)
    elg_error();
  entry->seqno = counter->count;
  entry->comid = counter->comid;

  /* RMA window not yet defined? */
  if (epk_vector_find(data->wins, entry,
                      (epk_vec_cmp_f)&double_id_cmp, &global_id) == 0) {
    /* Determine global ID */
    global_id = epk_vector_size(data->wins);

    epk_vector_push_back(data->wins, entry);
  } else {
    free(entry);
  }

  /* Add mapping */
  epk_idmap_add(data->maps[ESD_UNIFY_WINMAP], wid, global_id);
}

/*-------------------------------------------------------------------------*/
/**
 * @name Bit vector encoding
 * @{ */
/*-------------------------------------------------------------------------*/

/**
 * @brief        Count number of bits in a 32-bit value
 * @param  value 4-byte integer to be tested
 * @return       number of bits set in value
 */
elg_ui1 esd_count_bits(elg_ui4 value)
{
  return (BitSetTable[ value        & 0xff] +
          BitSetTable[(value >>  8) & 0xff] +
          BitSetTable[(value >> 16) & 0xff] +
          BitSetTable[(value >> 24) & 0xff]);
}

/**
 * @brief Count the number of bits set in the complete buffer
 *
 * @param  grpc number of blocks of sizeof(elg_ui1) in grpv
 * @param  grpv pointer to buffer
 * @return      number of bits set in grpv
 */
elg_ui4 esd_count_bits_in_buf(const elg_ui4 grpc, const elg_ui1* grpv)
{
  elg_ui4  count  = 0;              /* number of members in group */
  elg_ui4  resid  = 0;              /* tail of grpv */
  elg_ui4* p      = (elg_ui4*)grpv; /* convenience pointer to grgv */
  elg_ui4  i;                       /* loop variable */

  /* count bits in chunks of elg_ui4 (32-bit) */
  for (i = 0; i < (grpc/4); i++)
  {
    count += esd_count_bits(p[i]);
  }
  /* count bits in residue */
  memcpy(&resid, &p[i], (grpc % 4) * sizeof(elg_ui1));
  count += esd_count_bits(resid);

  return count;
}

/**
 * @brief Encode group 
 *
 * Encode process group with one of the built-in encoding schemes.
 *
 * @param  data Pointer to unification data structure
 * @param  grpc Number of blocks in incoming bitvector
 * @param  grpv Pointer to the incoming bitvector
 * @return      Pointer to group representation struct
 */
UnifyGroupRep* esd_bitvec_encode(EsdUnifyData*  data,
                                 const elg_ui4  grpc, 
                                 const elg_ui1* grpv)
{
  UnifyGroupRep* grprep  = (UnifyGroupRep*)malloc(sizeof(UnifyGroupRep));
  elg_ui4        count   = esd_count_bits_in_buf(grpc, grpv);

  assert(data);

  /* First communicator is MPI_COMM_WORLD, implicitly defining the
   * number of processes */
  if (data->tmp_max_procs == 0)
  {
    data->tmp_max_procs = count;
    elg_cntl_msg("MPI_COMM_WORLD contains %i processes", 
                 data->tmp_max_procs);
  }
  elg_cntl_msg("Current communicator contains %i processes", count);

  if (!grprep)
    elg_error();

  /* test if number of set bits is less than 1/32 */
  if (count * sizeof(elg_ui4) < grpc)
  {
    elg_cntl_msg("Preparing for sparse encoding");
    /* sparse representation */
    grprep->rep  = ESD_GROUP_SPARSE;
    grprep->size = count * sizeof(elg_ui4);
    grprep->buf  = (elg_ui1*) malloc(grprep->size);
    if (!grprep->buf)
      elg_error();

    esd_bitvec_sparse_encode(data, grpc, grpv, grprep);
  }
  /* test if number of set bits is greater than 31/32 */
  else if ((data->tmp_max_procs - count) * sizeof(elg_ui4) < grpc)
  {
    /* inverse sparse representation */
    grprep->rep  = ESD_GROUP_INV_SPARSE;
    grprep->size = (data->tmp_max_procs - count) * sizeof(elg_ui4);
    grprep->buf  = (grprep->size > 0) ? (elg_ui1*)malloc(grprep->size) : NULL;
    if (grprep->size && !grprep->buf)
      elg_error();

    esd_bitvec_inv_sparse_encode(data, grpc, grpv, grprep);
  }
  else
  {
    /*
     * Encode with RLE using plain bitvector as fallback. Due to the
     * fallback, all allocations and the setting of the encoding used
     * has to be done within the call.
     */
    esd_bitvec_rle_encode(data, grpc, grpv, grprep);
  }

  return grprep;
}

void esd_bitvec_decode(EsdUnifyData*        data,
                       elg_ui4              outc,
                       elg_ui1*             outv,
                       const UnifyGroupRep* grprep)
{
  switch (grprep->rep)
  {
    case ESD_GROUP_BITVECTOR:
    {
      memcpy(outv, grprep->buf, outc);
      break;
    }
    case ESD_GROUP_SPARSE:
    {
      esd_bitvec_sparse_decode(data, outc, outv, grprep); 
      break;
    }
    case ESD_GROUP_INV_SPARSE:
    {
      esd_bitvec_inv_sparse_decode(data, outc, outv, grprep);
      break;
    }
    case ESD_GROUP_RLE:
    {
      esd_bitvec_rle_decode(data, outc, outv, grprep);
      break;
    }
    default:
    {
      elg_error_msg("Unknown group compression\n");
      break;
    }

  }
}


/**
 * Encode a given bitvector as ascending sparse list of ranks
 */
void esd_bitvec_sparse_encode(EsdUnifyData*  data,
                              const elg_ui4  grpc, 
                              const elg_ui1* grpv,
                              UnifyGroupRep* grprep)
{
  elg_ui4 idx = 0;
  elg_ui4 i   = 0;

  assert(data);
  assert(grprep);

  /*
   * Header is already set by calling function, thus only the
   * buffer needs to be filled.
   */

  while ((i < data->tmp_max_procs) 
         && (idx < grprep->size / sizeof(elg_ui4)))
  {
    if (grpv[i/8] & (1 << (i % 8)))
    {
      elg_cntl_msg("Adding rank %i to sparse list", i);
      /* outv needs to be casted */
      ((elg_ui4*)(grprep->buf))[idx] = i;
      idx++;
    }
    i++;
  }
}

/**
 * Decode a given bitvector from ascending sparse list of ranks
 */
void esd_bitvec_sparse_decode(EsdUnifyData*        data,
                              elg_ui4              grpc,
                              elg_ui1*             grpv,
                              const UnifyGroupRep* grprep)
{
  elg_ui4  i      = 0;
  elg_ui4  p_size = grprep->size / sizeof(elg_ui4);
  elg_ui4* p      = (elg_ui4*)(grprep->buf);

  /* initialize bitvector */
  memset(grpv, 0, grpc);

  /* set member bits */
  for (i=0; i < p_size; i++)
  {
    elg_cntl_msg("Setting appropriate bit for rank %i", p[i]);
    grpv[p[i] / 8] |= (1 << (p[i] % 8));
  }
}

/**
 * Encode a given bitvector as ascending inverse sparse list of ranks
 */

void esd_bitvec_inv_sparse_encode(EsdUnifyData*  data,
                                  const elg_ui4  grpc, 
                                  const elg_ui1* grpv,
                                  UnifyGroupRep* grprep)
{
  elg_ui4 idx     = 0;
  elg_ui4 i       = 0;

  /*
   * Header is already set by calling function, thus only the
   * buffer needs to be filled.
   */

  while ((i < data->tmp_max_procs) && (idx < grprep->size / sizeof(elg_ui4)))
  {
    if (!(grpv[i/8] & (1 << (i % 8))))
    {
      elg_cntl_msg("Adding rank %i to inverse sparse list", i);
      /* outv needs to be casted */
      ((elg_ui4*)grprep->buf)[idx] = i;
      idx++;
    }
    i++;
  }
}


/**
 * Decode a given bitvector from ascending inverse sparse list of ranks
 */
void esd_bitvec_inv_sparse_decode(EsdUnifyData*        data,
                                  elg_ui4              grpc,
                                  elg_ui1*             grpv,
                                  const UnifyGroupRep* grprep)
{
  elg_ui4  i      = 0;
  elg_ui4  p_size = grprep->size / sizeof(elg_ui4);
  elg_ui4* p      = (elg_ui4*)(grprep->buf);

  /* initialize bitvector */
  memset(grpv, 255, grpc);

  /* unset non-member bits */
  for (i=0; i < p_size; i++)
  {
    elg_cntl_msg("Setting appropriate bit for rank %i", p[i]);
    grpv[p[i] / 8] &= ~(1 << (p[i] % 8));
  }
  /* unset unused bits */
  for (i = data->tmp_max_procs; i < grpc * 8; i++)
  {
    grpv[i / 8] &= ~(1 << (i % 8));
  }
}

/**
 * @brief Encode a given bitvector as runlength encoded byte-stream
 *
 * A given byte stream is RLE compressed using a threshold algorithm. If
 * any byte in the stream is repeated twice, the count of additional
 * repetitions of this byte is appended to these bytes. This can, in
 * cases where there are a lot of two-byte sequences lead to a larger
 * stream than the uncompressed stream, but should yield good results in
 * the mean.
 *
 * @note   As this function uses the uncompressed bitvector as a fallback
 *         in case the RLE compression would use more space, the 
 *         compression type and size cannot be determined in advance!
 */
void esd_bitvec_rle_encode(EsdUnifyData*  data,
                           const elg_ui4  grpc, 
                           const elg_ui1* grpv,
                           UnifyGroupRep* grprep)
{
  /* current index into incoming bitvector */
  elg_ui4        idx      = 0;
  /* counter for run-length of an RLE block. It counts the additional
   * bytes of the same pattern. Thus a run can have a maximum length of
   * 257 bytes. */
  elg_ui1        count    = 0;
  /* pointer to first element of each comparison */
  const elg_ui1* first    = grpv;
  /* pointer to second of each comparison */
  const elg_ui1* second   = grpv + 1;
  /* Indicator for overflow of RLE compression vector */
  int            overflow = 0;

  assert(data);
  assert(data->tmp_bitvec);
  assert(grprep);

  /* only perform RLE compression, if incoming bitvector is at least 3
   * bytes long. */
  if (grpc > 3)
  {
    elg_cntl_msg("Starting RLE compression");

    /*----------------------------------------------------------------------*/
    /* check for runs and compress */
    /*----------------------------------------------------------------------*/

    /* compress until either
     * a) RLE compression would result in a larger buffer, or
     * b) the end of the input buffer is reached.
     */
    while (!overflow && (second < (grpv + grpc)))
    {
      /* if @a first and @a second point to equal values, start counting
       * the runlength */
      if (*first == *second)
      {
        elg_cntl_msg("Run detected");
        /* no additional bytes so far */
        count = 0;
        /* this run will need at least 3 bytes. If this goes beyond the
         * incoming byte count, the overflow limit is reached */
        if ((idx + 3) > grpc)
        {
          overflow = 1;
          break;
        }
        else
        {
          /* save the run to the temporary buffer */
          data->tmp_bitvec[idx++] = *first;   /* save first byte */
          data->tmp_bitvec[idx++] = *second;  /* save second byte */
          while (((++second) < (grpv + grpc))  /* check for repetition */
                 && (*second == *first)
                 && (count < 255))
          {
            count++;
          }
          data->tmp_bitvec[idx++] = count;    /* save count */
          elg_cntl_msg("Saved run to buffer: (%i, %i, %i)",
                       *first, *second, count);

          /* done with this run -> set up encoding the rest */

          first = second;   /* set new start */
          second++;

          /* 
           * if second is now set behind the end of grpv, we need to
           * save the last byte, as the loop will be exited after this
           * iteration
           */
          if (second == (grpv + grpc))
          {
            if (idx + 1 > grpc)
            {
              overflow = 1;
              break;
            }
            else
            {
              elg_cntl_msg("Saved remaining byte to buffer: %i", *first);
              data->tmp_bitvec[idx++] = *first;
            }
          }
        }
      }
      else
      {
        /* check for overflow of temporary buffer */
        if ((idx + 1) > grpc) 
        {
          overflow = 1;
          break;
        }
        else
        {
          elg_cntl_msg("Saved byte to buffer: %i", *first);
          data->tmp_bitvec[idx++] = *first;  /* save first byte */

          first = second;
          second++;

          /* 
           * if second is now set behind the end of grpv, we need to
           * save the last byte, as the loop will be exited after this
           * iteration
           */
          if (second == (grpv + grpc))
          {
            if (idx + 1 > grpc)
            {
              overflow = 1;
              break;
            }
            else
            {
              elg_cntl_msg("Saved remaining byte to buffer: %i", *first);
              data->tmp_bitvec[idx++] = *first;
            }
          }
        }
      }
    }

    /* check, whether compression failed */
    if (!overflow)
    {
      elg_cntl_msg("Compression completed");
      /* save RLE compressed buffer to data structure */
      grprep->rep  = ESD_GROUP_RLE;
      grprep->size = idx * sizeof(elg_ui1);
      grprep->buf  = (elg_ui1*) malloc(grprep->size);
      if (!grprep->buf)
      {
        elg_error();
      }
      memcpy(grprep->buf, data->tmp_bitvec, grprep->size);
      elg_cntl_msg("Buffer size is now %i bytes (compression ratio %lf)",
                   grprep->size, grprep->size/grpc);
      
      return;
    }
  }

  /* if we get to this point, we have to fallback on plain bitvector */

  /* fallback to no-compression, i.e. use plain bitvector */
  elg_cntl_msg("Compression aborted");
  grprep->rep  = ESD_GROUP_BITVECTOR;
  grprep->size = grpc * sizeof(elg_ui1);
  grprep->buf  = (elg_ui1*) malloc(grprep->size);
  if (!grprep->buf)
  {
    elg_error();
  }
  memcpy(grprep->buf, grpv, grprep->size);
  elg_cntl_msg("Original buffer saved");
}

/**
 * Decode a given bitvector from runlength encoded byte-stream
 */
void esd_bitvec_rle_decode(EsdUnifyData*        data,
                           elg_ui4              grpc,
                           elg_ui1*             grpv,
                           const UnifyGroupRep* grprep)
{
  elg_ui4  idx    = 0;
  elg_ui1  count  = 0;
  elg_ui1* first;
  elg_ui1* second;

  elg_cntl_msg("Staring RLE decompression");

  /* compressed buffers are at least 3 bytes long */
  first  = grprep->buf;
  second = grprep->buf + 1;
  
  elg_cntl_msg("Saved byte %i", *first);
  grpv[idx++] = *first;
  while ((second - grprep->buf) < grprep->size)
  {
    elg_cntl_msg("Saved byte %i", *second);
    grpv[idx++] = *second;
    if (*first == *second)
    {
      elg_cntl_msg("Detected run. Additional bytes %i",
                   count);
      count = *(second+1);
      while (count--)
      {
        elg_cntl_msg("Saved byte %i", *second);
        grpv[idx++] = *second;
      } 
      elg_cntl_msg("Resetting pointers to compressed buffer");

      /* restart after count */
      first = second + 2;
      if ((first - grprep->buf) < grprep->size)
      {
        elg_cntl_msg("Saved byte %i", *second);
        grpv[idx++] = *first;
      }
      second = first + 1;
    }
    else
    {
      first = second;
      second++;
    }
  }

  elg_cntl_msg("RLE Decrompression completed");
}

/**
 * @}
 */

/* Include empty callback functions for event records */
#include "elg_readcb_dummies.c.gen"
