/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
**  KOJAK       http://www.fz-juelich.de/jsc/kojak/                        **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2003-2008                                                **
**  University of Tennessee, Innovative Computing Laboratory               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/

#ifndef _ESD_UNIFY_H
#define _ESD_UNIFY_H


#include <stddef.h>
#include <stdio.h>

#include "elg_rw.h"
#include "elg_impl.h"
#include "epk_idmap.h"
#include "epk_vector.h"
#include "epk_hashtab.h"


#ifdef __cplusplus
#  define EXTERN extern "C"
#else
#  define EXTERN extern 
#endif


/*--- Type definitions ----------------------------------------------------*/

/** Opaque data structure storing global and per-process unification data. */
typedef struct EsdUnifyData_struct EsdUnifyData;

/**
 * Enumeration type defining symbolic names for the different per-process
 * identifier mapping tables handled by EsdUnifyData.
 */
typedef enum {
  ESD_UNIFY_STRMAP,       /**< String ID mapping table */
  ESD_UNIFY_MACHMAP,      /**< Machine ID mapping table */
  ESD_UNIFY_NODEMAP,      /**< Node ID mapping table */
  ESD_UNIFY_LOCMAP,       /**< Location ID mapping table */
  ESD_UNIFY_FILEMAP,      /**< File ID mapping table */
  ESD_UNIFY_REGMAP,       /**< Region ID mapping table */
  ESD_UNIFY_CSITEMAP,     /**< Call site ID mapping table */
  ESD_UNIFY_CPATHMAP,     /**< Call-path ID mapping table */
  ESD_UNIFY_COMMAP,       /**< MPI communicator ID mapping table */
  ESD_UNIFY_WINMAP,       /**< RMA window ID mapping table */
  ESD_UNIFY_NUM_MAPS      /**< Total number of mapping tables */
} esd_unify_map_t;

/**
 * Enumeration type defining type of group vector representation in
 * memory
 */
typedef enum {
  ESD_GROUP_BITVECTOR,    /**< Legacy bitvector representation */
  ESD_GROUP_SPARSE,       /**< Sparse represenation */
  ESD_GROUP_INV_SPARSE,   /**< Inverse sparse repesentation */
  ESD_GROUP_RLE           /**< Run length encoded representation */
} esd_group_rep_t;

/*
 *---------------------------------------------------------------------------
 * EsdUnify
 *---------------------------------------------------------------------------
 */

/** @name Unification data handling
 ** @{ */

/* Construction & destruction */
EXTERN EsdUnifyData* esd_unify_create();
EXTERN void          esd_unify_free  (EsdUnifyData* instance);

/* Unification */
EXTERN void          esd_unify_buffer(      EsdUnifyData* instance,
                                      const buffer_t      buffer,
                                            size_t        size,
                                            ElgOut*       mapfile);

/* Identifier mapping table management */
EXTERN EpkIdMap*     esd_unify_get_map   (const EsdUnifyData*   instance,
                                                esd_unify_map_t map_id);
EXTERN void          esd_unify_clear_maps(      EsdUnifyData*   instance);

/* File I/O */
EXTERN void          esd_unify_write_defs(EsdUnifyData* instance,
                                          ElgOut*       deffile);

/** @} */

/*--- Type definitions ----------------------------------------------------*/

/*
 * "Base class" for entities which can be uniquely identified using a single
 * identifier.
 */
typedef struct
{
  elg_ui4 id;     /* Identifier */
} UnifySingleId;

/*
 * "Base class" for entities which can only be uniquely identified using two
 * identifiers.
 */
typedef struct
{
  elg_ui4 id1;     /* First ID */
  elg_ui4 id2;     /* Second ID */
} UnifyDoubleId;

/*
 * Structure for storing the data of an unified machine record. It is
 * "derived" from UnifySingleId.
 */
typedef struct
{
  elg_ui4 id;          /* Local machine ID */
  elg_ui4 nodec;       /* Number of nodes */
  elg_ui4 nameid;      /* Global name string ID */
} UnifyMach;

/*
 * Structure for storing the data of an unified node record. It is
 * "derived" from UnifyDoubleId, since a node identifier is local
 * to a machine.
 */
typedef struct
{
  elg_ui4 id;          /* Local node ID */
  elg_ui4 machid;      /* Global machine ID */
  elg_ui4 cpuc;        /* Number of CPUs */
  elg_ui4 nameid;      /* Global name string ID */
  elg_d8  clockrt;     /* Clock rate */
} UnifyNode;

/*
 * Structure for storing the data of an unified process record. It is
 * "derived" from UnifySingleId.
 */
typedef struct
{
  elg_ui4 id;         /* Local process ID */
  elg_ui4 nameid;     /* Global name string ID */
} UnifyProc;

/*
 * Structure for storing the data of an unified thread record. It is
 * "derived" from UnifyDoubleId, since thread identifiers are local
 * to their corresponding process.
 */
typedef struct
{
  elg_ui4 id;         /* Local thread ID */
  elg_ui4 pid;        /* Local process ID */
  elg_ui4 nameid;     /* Global name string ID */
} UnifyThrd;

/* Structure for storing the data of an unified location record. */
typedef struct
{
  elg_ui4 machid;     /* Global machine ID */
  elg_ui4 nodeid;     /* Global node ID */
  elg_ui4 procid;     /* Local process ID */
  elg_ui4 thrdid;     /* Local thread ID */
} UnifyLoc;

/* Structure for storing the data of an unified source code region record. */
typedef struct
{
  elg_ui4 nameid;     /* Global name string ID */
  elg_ui4 fileid;     /* Global file ID */
  elg_ui4 begln;      /* Begin line number */
  elg_ui4 endln;      /* End line number */
  elg_ui4 descid;     /* Global description string ID */
  elg_ui1 type;       /* Region type */
} UnifyRegion;

/* Structure for storing the data of an unified call site record. */
typedef struct
{
  elg_ui4 fileid;      /* Global file ID */
  elg_ui4 lineno;      /* Line number */
  elg_ui4 e_regid;     /* Global ID of region to be entered */
  elg_ui4 l_regid;     /* Global ID of region to be left */
} UnifyCsite;

/* Structure for storing the data of an unified call-path record. */
typedef struct
{
  elg_ui4    rid;          /* Global region ID */
  elg_ui4    ppid;         /* Global ID of parent call-path */
  elg_ui4    id;           /* Global call-path ID (needed because records are stored in multiple lists) */
  elg_ui8    order;        /* Node order specifier */
  EpkVector* children;     /* List of child nodes */
} UnifyCpath;

/*
 * Structure for storing the data of an unified hardware counter metric
 * record. It is "derived" from UnifySingleId.
 */
typedef struct
{
  elg_ui4 id;         /* Global metric ID */
  elg_ui4 nameid;     /* Global name string ID */
  elg_ui4 descid;     /* Global description string ID */
  elg_ui1 type;       /* Metric type */
  elg_ui1 mode;       /* Metric mode */
  elg_ui1 ival;       /* Time interval */
} UnifyMetric;

/*
 * Group Representation structure
 */
typedef struct
{
  esd_group_rep_t rep;      /* Group encoding scheme */
  elg_ui4         size;     /* Size of memory buffer */
  elg_ui1*        buf;      /* Memory buffer storing encoded group rep */
} UnifyGroupRep;

/*
 * Structure for use as reference counted bitstrings
 */
typedef struct
{
  elg_ui4        refs; /* Counter for reference on this group */
  UnifyGroupRep* vec;  /* Bitvector defining a group of processes */
} UnifyGroup;

/* Structure for storing the data of an unified MPI communicator record. */
typedef struct
{
  elg_ui4     seqno;  /* Sequence number */
  UnifyGroup* group;  /* Group representation */
  elg_ui1     uag;    /* Flag indicating whether this communicator is used
                         as a group */
} UnifyCom;

/*
 * Structure for storing the data of an unified cartesian topology record.
 * It is "derived" from UnifySingleId.
 */ 
typedef struct
{
  elg_ui4    comid;      /* Global communicator ID */
  elg_ui4    ndims;      /* Number of dimensions */
  elg_ui4*   nlocs;      /* Number of locations per dimension */
  elg_ui1*   period;     /* Periodicity per location */
  EpkVector* coords;     /* Coordinates */
} UnifyCart;

/*
 * Structure for storing the data of an unified RMA window record. It is
 * "derived" from UnifyDoubleId.
 */
typedef struct
{
  elg_ui4 seqno;     /* Sequence number */
  elg_ui4 comid;     /* Global communicator ID */
} UnifyWin;

/*
 * Structure for storing the data of MPI groups and how often they are used
 * to define an MPI communicator.
 */
typedef struct
{
  UnifyGroup* group;    /* Group representation */
  elg_ui4     count;    /* Number of occurrences */
} GrpCount;

/*
 * Structure for storing how often an MPI communicator is used to define
 * an RMA window. It is "derived" from UnifySingleId.
 */
typedef struct
{
  elg_ui4 comid;     /* Global communicator ID */
  elg_ui4 count;     /* Number of occurrences */
} ComCount;

/*
 * Structure for storing the data of unified cartesian coordinate records.
 * It is "derived" from UnifySingleId.
 */
typedef struct
{
  elg_ui4  locid;     /* Local/global location ID */
  elg_ui4  topid;     /* Global topology ID */
  elg_ui4* coord;     /* Coordinate */
} CartCoord;


/*--- Type definitions ----------------------------------------------------*/

/* Actual data structure storing temporary and unification results. */
struct EsdUnifyData_struct
{
  /*--- Temporary data ---*/
  elg_ui4     tmp_id;         /* Temporary identifier */
  elg_ui4     tmp_count;      /* Temporary count */
  char*       tmp_string;     /* Temporary string buffer */
  EpkVector*  tmp_coord;      /* Temporary topology coordinate buffer */
  EpkVector*  grpcnt;         /* Temporary vector to count how many times a
                                 communicator is created for a certain group */
  EpkVector*  comcnt;         /* Temporary vector to count how many times a
                                 window is created for a certain communicator */

  elg_ui1*    tmp_bitvec;     /* temporary buffer for compression */
  elg_ui4     tmp_bitvec_size;/* size of temporary compression buffer */
  elg_ui4     tmp_max_procs;  /* number of total processes */


  /*--- Global unification data ---*/
  EpkHashTab* str2id;         /* String |-> Global string ID */
  EpkIdMap*   topmap;         /* Global communicator ID |-> Global topology ID */

  EpkVector*  strings;        /* Unified strings */
  EpkVector*  machs;          /* Unified machines */
  EpkVector*  nodes;          /* Unified nodes */
  EpkVector*  procs;          /* Unified processes */
  EpkVector*  thrds;          /* Unified threads */
  EpkVector*  locs;           /* Unified locations */
  EpkVector*  files;          /* Unified files */
  EpkVector*  regions;        /* Unified regions */
  EpkVector*  csites;         /* Unified call sites */
  EpkVector*  cpaths;         /* Unified call-paths */
  EpkVector*  cproots;        /* Root nodes of unified call-paths */
  EpkVector*  metrics;        /* Unified metrics */
  EpkVector*  coms;           /* Unified communicators */
  EpkVector*  tops;           /* Unified topologies */
  EpkVector*  wins;           /* Unified RMA windows */

  EpkVector*  mapoffs;        /* File offsets in mapping file */

  /*--- Per-process unification data ---*/
  EpkIdMap*   maps[ESD_UNIFY_NUM_MAPS];     /* Identifier mapping tables */

  elg_ui1     offcnt;         /* Number of time synchronization measurements */
  elg_d8      ltime[2];       /* Local time at start/end of measurement */
  elg_d8      offset[2];      /* Time offset at start/end of measurement */

  elg_ui4     ntypes;         /* Number of event types/counts */
  elg_ui1*    evtypes;        /* Array containing event types */
  elg_ui4*    evcounts;       /* Array containing event counts */
};


#endif   /* !_ESD_UNIFY_H */
