/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


#include <sstream>

#include <mpi.h>

#include "Error.h"
#include "GlobalDefs.h"
#include "MpiCartesian.h"
#include "MpiComm.h"
#include "MpiDefsFactory.h"
#include "MpiGroup.h"
#include "MpiWindow.h"
#include "Process.h"

using namespace std;
using namespace pearl;


/*
 *---------------------------------------------------------------------------
 *
 * class MpiDefsFactory
 *
 *---------------------------------------------------------------------------
 */

//--- MPI-specific factory methods ------------------------------------------

void MpiDefsFactory::createMpiGroup(GlobalDefs& defs,
                                    ident_t     id,
                                    uint32_t    size,
                                    uint8_t*    group) const
{
  // Convert group description
  process_group processes;
  convertProcessGroup(defs, size, group, processes);

  // Sanity check
  int world_size;
  MPI_Comm_size(MPI_COMM_WORLD, &world_size);
  if (static_cast<uint32_t>(world_size) < processes.size()) {
    ostringstream num;
    num << defs.num_processes();

    throw FatalError("MpiDefsFactory::createMpiGroup() -- "
                     "Experiment requires at least " + num.str() + " MPI ranks.");
  }

  // Create & store new MPI communicator object
  defs.add_group(new MpiGroup(id, processes));
}


void MpiDefsFactory::createMpiComm(GlobalDefs& defs,
                                   ident_t     id,
                                   uint32_t    size,
                                   uint8_t*    group) const
{
  // Convert group description
  process_group processes;
  convertProcessGroup(defs, size, group, processes);

  // Sanity check
  int world_size;
  MPI_Comm_size(MPI_COMM_WORLD, &world_size);
  if (static_cast<uint32_t>(world_size) < processes.size()) {
    ostringstream num;
    num << defs.num_processes();

    throw FatalError("MpiDefsFactory::createMpiComm() -- "
                     "Experiment requires at least " + num.str() + " MPI ranks.");
  }

  // Create & store new MPI communicator object
  defs.add_group(new MpiComm(id, processes));
}


void MpiDefsFactory::createMpiWindow(GlobalDefs& defs,
                                     ident_t     id,
                                     ident_t     comm_id) const
{
  // Determine communicator
  MpiComm* comm = dynamic_cast<MpiComm*>(defs.get_group(comm_id));
  if (!comm)
    throw RuntimeError("MpiDefsFactory::createMpiWindow() -- "
                       "Invalid MPI communicator ID.");

  // Create & store new MPI RMA window
  defs.add_window(new MpiWindow(id, comm));
}


void MpiDefsFactory::createMpiCartesian(GlobalDefs&        defs,
                                        ident_t            id,
                                        const cart_dims&   dimensions,
                                        const cart_period& periodic,
                                        ident_t            comm_id) const
{
  // Determine communicator
  MpiComm* comm = dynamic_cast<MpiComm*>(defs.get_group(comm_id));
  if (!comm)
    throw RuntimeError("MpiDefsFactory::createMpiCartesian() -- "
                       "Invalid MPI communicator ID.");

  // Create & store new MPI cartesian topology object
  defs.add_cartesian(new MpiCartesian(id, dimensions, periodic, comm));
}
