/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


#include <new>

#include <mpi.h>

#include <elg_error.h>

#include "Calltree.h"
#include "GlobalDefs.h"
#include "MpiDefsFactory.h"
#include "MpiEventFactory.h"
#include "MpiMessage.h"
#include "SmallObjAllocator.h"
#include "pearl.h"

using namespace std;
using namespace pearl;


//--- Function prototypes ---------------------------------------------------

namespace pearl
{

extern void PEARL_new_handler();

}   /* namespace pearl */


//--- Library initialization ------------------------------------------------

/**
 * Initializes the PEARL library and installs a custom out-of-memory handler.
 * It is required to call one of the PEARL initialization functions before
 * calling any other PEARL function or instantiating any PEARL class.
 *
 * @attention Make sure to initialize PEARL before installing any exit
 *            handlers using atexit(). Otherwise, you might get error
 *            messages about memory leaks.
 *
 * @note This function is intended to be used in pure MPI-based PEARL
 *       programs.
 *
 * @see PEARL_init(), PEARL_omp_init(), PEARL_hybrid_init()
 */
void pearl::PEARL_mpi_init()
{
  // Register factories & allocators
  DefsFactory::registerFactory(new MpiDefsFactory());
  EventFactory::registerFactory(new MpiEventFactory());
  SmallObjAllocator::registerAllocator(new SmallObjAllocator());

  // Register new handler (out of memory)
  set_new_handler(PEARL_new_handler);
}


//--- Call tree unification -------------------------------------------------

/**
 * Combines process-local call trees into a global one. Since this is
 * conceptually a sequential process, the actual unification is only carried
 * out if the process-local call trees have been modified after reading the
 * trace, i.e., the call tree provided by the global definitions object was
 * incomplete. Therefore, this function should only be called after verifying
 * the process-local call tree using PEARL_verify_calltree() or
 * PEARL_omp_verify_calltree(), but before preprocessing the trace data via
 * PEARL_preprocess_trace() or PEARL_omp_preprocess_trace().
 *
 * @note This function is intended to be used in pure MPI-based or hybrid
 *       OpenMP/MPI PEARL programs. In the case of a hybrid OpenMP/MPI
 *       program, make sure to call this function <b>only from the master
 *       thread.</b>
 *
 * @param defs Global definitions object
 *
 * @see PEARL_verify_calltree(), PEARL_omp_verify_calltree(),
 *      PEARL_preprocess_trace(), PEARL_omp_preprocess_trace()
 */
void pearl::PEARL_mpi_unify_calltree(GlobalDefs& defs)
{
  Calltree* ctree = defs.get_calltree();

  // Determine whether call tree unification is necessary
  int global_mod = 0;
  int local_mod  = ctree->get_modified();
  MPI_Allreduce(&local_mod, &global_mod, 1, MPI_INT, MPI_SUM, MPI_COMM_WORLD);
  if (global_mod == 0)
    return;

  // Call tree unification
  int rank;
  int size;
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);

  if (rank == 0) {
    // Display informational message
    elg_warning("\nExperiment contains incomplete/inconsistent definition of"
                "\ncall tree. Reconstructing global call tree from trace data.");

    // Receive and merge local call trees
    for (int i = 1; i < size; ++i) {
      MpiMessage message(MPI_COMM_WORLD);
      message.recv(i, 0);
      Calltree local_ctree(defs, message);
      ctree->merge(local_ctree);
    }
  } else {
    // Pack and send local call tree
    MpiMessage message(MPI_COMM_WORLD);
    ctree->pack(message);
    message.send(0, 0);
  }

  // Broadcast global call tree
  MpiMessage message(MPI_COMM_WORLD);
  ctree->pack(message);
  message.bcast(0);
  Calltree* global_ctree = new Calltree(defs, message);
  defs.set_calltree(global_ctree);
}
