/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


#include <algorithm>
#include <cassert>

#include "Error.h"
#include "Functors.h"
#include "Group.h"
#include "Process.h"

using namespace std;
using namespace pearl;


/*
 *---------------------------------------------------------------------------
 *
 * class Group
 *
 *---------------------------------------------------------------------------
 */

//--- Constructors & destructor ---------------------------------------------

/**
 * Creates a new instance and initializes the data with the given values
 * @a id and @a processes, respectively. The given @a processes have to be
 * sorted by their numerical identifiers (i.e., their rank numbers) in
 * ascending order!
 *
 * @param id        Group identifier
 * @param processes List of processes defining the group
 */
Group::Group(ident_t id, const process_group& processes)
  : IdObject(id),
    m_processes(processes)
{
#ifndef NDEBUG
  uint32_t count = processes.size();
  for (uint32_t i = 0; i < count; ++i)
    assert(processes[i]);
#endif
}


//--- Get group information -------------------------------------------------

/**
 * Returns the number of processes belonging to this group.
 *
 * @return Number of processes
 */
uint32_t Group::num_processes() const
{
  return m_processes.size();
}


/**
 * Returns the Process with the given @a rank with respect to this group.
 *
 * @param rank Rank number of the requested process
 *
 * @return Corresponding Process object
 *
 * @exception RuntimeError if an invalid @a rank is given ("Rank out of
 *                         range")
 */
Process* Group::get_process(uint32_t rank) const
{
  if (rank >= m_processes.size())
    throw RuntimeError("Group::get_process(uint32_t) -- Rank out of range.");

  return m_processes[rank];
}


/**
 * Returns the rank number of the given @a process with respect to this
 * group.
 *
 * @param process Process object
 *
 * @return Rank number
 *
 * @exception RuntimeError if the @a process is not part of this group
 *                         ("Unknown process")
 */
uint32_t Group::get_rank(Process* process) const
{
  assert(process);

  // Search process object
  process_group::const_iterator it;
  it = lower_bound(m_processes.begin(), m_processes.end(),
                   process, less_ptr<Process*>());

  // Found? => Return rank number
  if (it != m_processes.end() && !less_ptr<Process*>()(process, *it))
    return it - m_processes.begin();

  // Not found...
  throw RuntimeError("Group::get_rank(Process*) -- Unknown process.");
}
