/****************************************************************************
**  SCALASCA    http://www.scalasca.org/                                   **
*****************************************************************************
**  Copyright (c) 1998-2010                                                **
**  Forschungszentrum Juelich, Juelich Supercomputing Centre               **
**                                                                         **
**  Copyright (c) 2006-2010                                                **
**  TU Dresden, Zentrum fuer Informationsdienste und Hochleistungsrechnen  **
**                                                                         **
**  See the file COPYRIGHT in the package base directory for details       **
****************************************************************************/


#include <algorithm>
#include <cstdlib>
#include <stack>

#include <elg_error.h>
#include <elg_readcb.h>
#include <epk_archive.h>

#include "Calltree.h"
#include "CNode.h"
#include "Enter_rep.h"
#include "Error.h"
#include "GlobalDefs.h"
#include "LocalTrace.h"
#include "MappingTable.h"
#include "readcb.h"

using namespace std;
using namespace pearl;


/*
 *---------------------------------------------------------------------------
 *
 * class LocalTrace
 *
 *---------------------------------------------------------------------------
 */

//--- Constructors & destructor ---------------------------------------------

LocalTrace::LocalTrace(const GlobalDefs& defs,
                       const string&     dirname,
                       uint32_t          rank,
                       uint32_t          thread_id)
  : m_defs(defs),
    m_location(NULL)
{
  // Create mapping table
  MappingTable* table = new MappingTable(defs, dirname, rank);

  // Determine file name
  char*  tmp_str  = epk_archive_rankname(EPK_TYPE_ELG, dirname.c_str(), rank);
  string filename = tmp_str;
  free(tmp_str);

  // Open local event trace file
  ElgRCB* file = elg_read_open(filename.c_str());
  if (!file)
    throw RuntimeError("Cannot open EPILOG event trace file \"" +
                       filename + "\".");

  // Read data
  TraceCbData data(defs, *this, *table, thread_id);
  while (elg_read_next_event(file, &data))
    ;
  elg_read_close(file);

  // Free mapping table
  delete table;

  // Check for errors
  if (!data.m_message.empty())
    throw Error(data.m_message);

  // Retrieve location
  m_location = defs.get_location(data.m_location);

  // Ensure consistency of ENTER_TRACING/EXIT_TRACING records
  preprocess();
}


LocalTrace::~LocalTrace()
{
  /* Delete events */
  container_type::reverse_iterator it = m_events.rbegin();
  while (it != m_events.rend()) {
    delete *it;
    ++it;
  }
}


//--- Iterator handling -----------------------------------------------------

LocalTrace::iterator LocalTrace::begin() const
{
  return Event(this);
}

LocalTrace::iterator LocalTrace::end() const
{
  return Event(this, true);
}


LocalTrace::reverse_iterator LocalTrace::rbegin() const
{
  return reverse_iterator(end());
}


LocalTrace::reverse_iterator LocalTrace::rend() const
{
  return reverse_iterator(begin());
}



//--- Get trace information -------------------------------------------------

uint32_t LocalTrace::size() const
{
  return m_events.size();
}


uint32_t LocalTrace::num_events() const
{
  return m_events.size();
}


Location* LocalTrace::get_location() const
{
  return m_location;
}


const GlobalDefs& LocalTrace::get_definitions() const
{
  return m_defs;
}


//--- Store new events ------------------------------------------------------

void LocalTrace::add_event(Event_rep* event)
{
  m_events.push_back(event);
}


//--- Private methods -------------------------------------------------------

Event_rep* LocalTrace::operator[](uint32_t index) const
{
  return m_events[index];
}


void LocalTrace::preprocess()
{
  // Ensure consistency for EXIT_TRACING/EXIT event combinations
  //   NOTE: This consistency check should eventually be carried out by the
  //         measurement system.
  uint32_t count = m_events.size();
  for (uint32_t i = 1; i < count; ++i) {
    if (m_events[i-1]->is_typeof(EXIT) && m_events[i]->is_typeof(EXIT) &&
        m_events[i-1]->get_time()      >  m_events[i]->get_time()) {

      std::swap(m_events[i-1], m_events[i]);

      for (uint32_t depth = 1, jj = i-2; depth; --jj) {
        if (m_events[jj]->is_typeof(EXIT))
          ++depth;
        if (m_events[jj]->is_typeof(ENTER))
          --depth;

        std::swap(m_events[jj], m_events[jj+1]);
      }
    }
  }
}
